# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import ENTITY_TYPE, recursive_tile
from ...serialization.serializables import BoolField, AnyField, KeyField, Int32Field
from ...tensor.utils import filter_inputs
from ..core import SERIES_TYPE, DATAFRAME_TYPE
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import build_empty_df, validate_axis, parse_index


class DataFrameCorr(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.CORR

    _other = KeyField("other")
    _method = AnyField("method")
    _min_periods = Int32Field("min_periods")
    _axis = Int32Field("axis")
    _drop = BoolField("drop")

    def __init__(
        self, other=None, method=None, min_periods=None, axis=None, drop=None, **kw
    ):
        super().__init__(
            _other=other,
            _method=method,
            _min_periods=min_periods,
            _axis=axis,
            _drop=drop,
            **kw,
        )

    @property
    def other(self):
        return self._other

    @property
    def method(self):
        return self._method

    @property
    def min_periods(self):
        return self._min_periods

    @property
    def axis(self):
        return self._axis

    @property
    def drop(self):
        return self._drop

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        next(inputs_iter)
        if isinstance(self._other, ENTITY_TYPE):
            self._other = next(inputs_iter)

    def __call__(self, df_or_series):
        if isinstance(df_or_series, SERIES_TYPE):
            inputs = filter_inputs([df_or_series, self.other])
            return self.new_scalar(inputs, dtype=np.dtype(np.float_))
        else:

            def _filter_numeric(obj):
                if not isinstance(obj, DATAFRAME_TYPE):
                    return obj
                num_dtypes = build_empty_df(obj.dtypes)._get_numeric_data().dtypes
                if len(num_dtypes) != len(obj.dtypes):
                    return obj[list(num_dtypes.index)]
                return obj

            df_or_series = _filter_numeric(df_or_series)
            self._other = _filter_numeric(self._other)

            inputs = filter_inputs([df_or_series, self.other])
            if self.axis is None:
                dtypes = pd.Series(
                    [np.dtype(np.float_)] * len(df_or_series.dtypes),
                    index=df_or_series.dtypes.index,
                )
                return self.new_dataframe(
                    inputs,
                    shape=(df_or_series.shape[1],) * 2,
                    dtypes=dtypes,
                    index_value=df_or_series.columns_value,
                    columns_value=df_or_series.columns_value,
                )
            else:
                new_index_value = df_or_series.axes[1 - self.axis].index_value
                if isinstance(self.other, DATAFRAME_TYPE):
                    align_dtypes = pd.concat(
                        [self.other.dtypes, df_or_series.dtypes], axis=1
                    )
                    align_shape = (np.nan, align_dtypes.shape[0])
                    new_index_value = parse_index(align_dtypes.index)
                else:
                    align_shape = df_or_series.shape

                shape = (np.nan,) if self.drop else (align_shape[1 - self.axis],)
                return self.new_series(
                    inputs,
                    shape=shape,
                    dtype=np.dtype(np.float_),
                    index_value=new_index_value,
                )

    @classmethod
    def _tile_single(cls, op: "DataFrameCorr"):
        out = op.outputs[0]

        new_op = op.copy().reset_key()
        chunk = new_op.new_chunk(
            [inp.chunks[0] for inp in op.inputs],
            index=(0,) * len(out.shape),
            **out.params,
        )

        new_op = op.copy().reset_key()
        return new_op.new_tileables(
            op.inputs, chunks=[chunk], nsplits=((s,) for s in out.shape), **out.params
        )

    @staticmethod
    def _tile_pearson_cross(left, right, min_periods):
        left_tensor, right_tensor = (
            left.fillna(0).to_tensor(),
            right.fillna(0).to_tensor(),
        )

        nna_left = left.notna().to_tensor().astype(np.float_)
        nna_right = right.notna().to_tensor().astype(np.float_)

        sum_left = left_tensor.T.dot(nna_right)
        sum_right = right_tensor.T.dot(nna_left)
        sum_left2 = (left_tensor.T**2).dot(nna_right)
        sum_right2 = (right_tensor.T**2).dot(nna_left)
        sum_mul = left_tensor.T.dot(right_tensor)
        data_count = nna_left.T.dot(nna_right)

        divisor = np.sqrt(data_count * sum_left2 - sum_left * sum_left).T * np.sqrt(
            data_count * sum_right2 - sum_right * sum_right
        )

        result = (data_count * sum_mul - sum_left * sum_right.T) / divisor
        if min_periods is not None:
            result = np.where(data_count >= min_periods, result, np.nan)
        return result

    @classmethod
    def _tile_pearson_align(cls, left, right, axis):
        nna_left = left.notna().astype(np.float_)
        nna_right = right.notna().astype(np.float_)

        left, right = left.fillna(0), right.fillna(0)

        sum_left = left.mul(nna_right, axis=axis).sum(axis=axis)
        sum_right = nna_left.mul(right, axis=axis).sum(axis=axis)
        sum_left2 = (left**2).mul(nna_right, axis=axis).sum(axis=axis)
        sum_right2 = nna_left.mul(right**2, axis=axis).sum(axis=axis)
        sum_mul = left.mul(right, axis=axis).sum(axis=axis)
        data_count = nna_left.mul(nna_right, axis=axis).sum(axis=axis)

        divisor = np.sqrt(data_count * sum_left2 - sum_left * sum_left) * np.sqrt(
            data_count * sum_right2 - sum_right * sum_right
        )
        return (data_count * sum_mul - sum_left * sum_right) / divisor

    @classmethod
    def _tile_series(cls, op: "DataFrameCorr"):
        left = op.inputs[0]
        right = op.other

        _check_supported_methods(op.method)
        return [
            (
                yield from recursive_tile(
                    cls._tile_pearson_cross(left, right, min_periods=op.min_periods)
                )
            )
        ]

    @classmethod
    def _tile_dataframe_cross(cls, op: "DataFrameCorr"):
        from ..initializer import DataFrame as MarsDataFrame

        left = op.inputs[0]
        right = op.other if op.other is not None else op.inputs[0]

        _check_supported_methods(op.method)

        result = cls._tile_pearson_cross(left, right, min_periods=op.min_periods)
        result = MarsDataFrame(
            result, index=left.dtypes.index, columns=right.dtypes.index
        )
        return [(yield from recursive_tile(result))]

    @classmethod
    def _tile_dataframe_align(cls, op: "DataFrameCorr"):
        left = op.inputs[0]
        right = op.other

        _check_supported_methods(op.method)
        result = cls._tile_pearson_align(left, right, axis=op.axis)
        if op.drop:
            result = result.dropna(axis=op.axis)
        return [(yield from recursive_tile(result))]

    @classmethod
    def tile(cls, op: "DataFrameCorr"):
        inp = op.inputs[0]
        if len(inp.chunks) == 1 and (op.other is None or len(op.other.chunks) == 1):
            return cls._tile_single(op)
        elif isinstance(inp, SERIES_TYPE):
            return (yield from cls._tile_series(op))
        elif op.axis is None:
            return (yield from cls._tile_dataframe_cross(op))
        else:
            return (yield from cls._tile_dataframe_align(op))

    @classmethod
    def execute(cls, ctx, op: "DataFrameCorr"):
        inp = op.inputs[0]
        out = op.outputs[0]
        inp_data = ctx[inp.key]

        if inp.ndim == 1:
            ctx[out.key] = inp_data.corr(
                ctx[op.other.key], method=op.method, min_periods=op.min_periods
            )
        elif op.axis is None:
            ctx[out.key] = inp_data.corr(method=op.method, min_periods=op.min_periods)
        else:
            ctx[out.key] = inp_data.corrwith(
                ctx[op.other.key], method=op.method, axis=op.axis, drop=op.drop
            )


def _check_supported_methods(method):
    if method != "pearson":
        raise NotImplementedError(f"Correlation method {method!r} not supported")


def df_corr(df, method="pearson", min_periods=1):
    """
    Compute pairwise correlation of columns, excluding NA/null values.

    Parameters
    ----------
    method : {'pearson', 'kendall', 'spearman'} or callable
        Method of correlation:

        * pearson : standard correlation coefficient
        * kendall : Kendall Tau correlation coefficient
        * spearman : Spearman rank correlation
        * callable: callable with input two 1d ndarrays
            and returning a float. Note that the returned matrix from corr
            will have 1 along the diagonals and will be symmetric
            regardless of the callable's behavior.

        .. note::
            kendall, spearman and callables not supported on multiple chunks yet.

    min_periods : int, optional
        Minimum number of observations required per pair of columns
        to have a valid result. Currently only available for Pearson
        and Spearman correlation.

    Returns
    -------
    DataFrame
        Correlation matrix.

    See Also
    --------
    DataFrame.corrwith : Compute pairwise correlation with another
        DataFrame or Series.
    Series.corr : Compute the correlation between two Series.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> df = md.DataFrame([(.2, .3), (.0, .6), (.6, .0), (.2, .1)],
    ...                   columns=['dogs', 'cats'])
    >>> df.corr(method='pearson').execute()
              dogs      cats
    dogs  1.000000 -0.851064
    cats -0.851064  1.000000
    """
    op = DataFrameCorr(method=method, min_periods=min_periods)
    return op(df)


def df_corrwith(df, other, axis=0, drop=False, method="pearson"):
    """
    Compute pairwise correlation.

    Pairwise correlation is computed between rows or columns of
    DataFrame with rows or columns of Series or DataFrame. DataFrames
    are first aligned along both axes before computing the
    correlations.

    Parameters
    ----------
    other : DataFrame, Series
        Object with which to compute correlations.
    axis : {0 or 'index', 1 or 'columns'}, default 0
        The axis to use. 0 or 'index' to compute column-wise, 1 or 'columns' for
        row-wise.
    drop : bool, default False
        Drop missing indices from result.
    method : {'pearson', 'kendall', 'spearman'} or callable
        Method of correlation:

        * pearson : standard correlation coefficient
        * kendall : Kendall Tau correlation coefficient
        * spearman : Spearman rank correlation
        * callable: callable with input two 1d ndarrays
            and returning a float.

        .. note::
            kendall, spearman and callables not supported on multiple chunks yet.

    Returns
    -------
    Series
        Pairwise correlations.

    See Also
    --------
    DataFrame.corr : Compute pairwise correlation of columns.
    """
    axis = validate_axis(axis, df)
    if drop:
        # TODO implement with df.align(method='inner')
        raise NotImplementedError("drop=True not implemented")
    op = DataFrameCorr(other=other, method=method, axis=axis, drop=drop)
    return op(df)


def series_corr(series, other, method="pearson", min_periods=None):
    """
    Compute correlation with `other` Series, excluding missing values.

    Parameters
    ----------
    other : Series
        Series with which to compute the correlation.
    method : {'pearson', 'kendall', 'spearman'} or callable
        Method used to compute correlation:

        - pearson : Standard correlation coefficient
        - kendall : Kendall Tau correlation coefficient
        - spearman : Spearman rank correlation
        - callable: Callable with input two 1d ndarrays and returning a float.

        .. note::
            kendall, spearman and callables not supported on multiple chunks yet.

    min_periods : int, optional
        Minimum number of observations needed to have a valid result.

    Returns
    -------
    float
        Correlation with other.

    See Also
    --------
    DataFrame.corr : Compute pairwise correlation between columns.
    DataFrame.corrwith : Compute pairwise correlation with another
        DataFrame or Series.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> s1 = md.Series([.2, .0, .6, .2])
    >>> s2 = md.Series([.3, .6, .0, .1])
    >>> s1.corr(s2, method='pearson').execute()
    -0.8510644963469898
    """
    op = DataFrameCorr(other=other, method=method, min_periods=min_periods)
    return op(series)


def series_autocorr(series, lag=1):
    """
    Compute the lag-N autocorrelation.

    This method computes the Pearson correlation between
    the Series and its shifted self.

    Parameters
    ----------
    lag : int, default 1
        Number of lags to apply before performing autocorrelation.

    Returns
    -------
    float
        The Pearson correlation between self and self.shift(lag).

    See Also
    --------
    Series.corr : Compute the correlation between two Series.
    Series.shift : Shift index by desired number of periods.
    DataFrame.corr : Compute pairwise correlation of columns.
    DataFrame.corrwith : Compute pairwise correlation between rows or
        columns of two DataFrame objects.

    Notes
    -----
    If the Pearson correlation is not well defined return 'NaN'.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> s = md.Series([0.25, 0.5, 0.2, -0.05])
    >>> s.autocorr().execute()  # doctest: +ELLIPSIS.execute()
    0.10355...
    >>> s.autocorr(lag=2).execute()  # doctest: +ELLIPSIS.execute()
    -0.99999...

    If the Pearson correlation is not well defined, then 'NaN' is returned.

    >>> s = md.Series([1, 0, 0, 0])
    >>> s.autocorr().execute()
    nan
    """
    op = DataFrameCorr(other=series.shift(lag), method="pearson")
    return op(series)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2017 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import binascii
import operator
import sys
import tokenize
import textwrap
from collections import OrderedDict
from functools import reduce
from io import StringIO

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import ENTITY_TYPE, OutputType, get_output_types, recursive_tile
from ...serialization.serializables import BoolField, DictField, StringField
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index

LOCAL_TAG = "_local_var_"
BACKTICK_TAG = "_backtick_var_"


def _tokenize_str(reader):
    token_generator = tokenize.generate_tokens(reader)

    def _iter_backtick_string(gen, line, back_start):
        for _, tokval, start, _, _ in gen:
            if tokval == "`":
                return (
                    BACKTICK_TAG
                    + binascii.b2a_hex(
                        line[back_start[1] + 1 : start[1]].encode()
                    ).decode()
                )
        else:
            raise SyntaxError(f"backtick quote at {back_start} does not match")

    for toknum, tokval, start, _, line in token_generator:
        if toknum == tokenize.OP:
            if tokval == "@":
                tokval = LOCAL_TAG
            if tokval == "&":
                toknum = tokenize.NAME
                tokval = "and"
            elif tokval == "|":
                toknum = tokenize.NAME
                tokval = "or"
        elif tokval == "`":
            yield tokenize.NAME, _iter_backtick_string(token_generator, line, start)
            continue
        yield toknum, tokval


class CollectionVisitor(ast.NodeVisitor):
    _op_handlers = {
        ast.Add: operator.add,
        ast.Sub: operator.sub,
        ast.Mult: operator.mul,
        ast.Div: operator.truediv,
        ast.FloorDiv: operator.floordiv,
        ast.mod: operator.mod,
        ast.Pow: operator.pow,
        ast.Eq: operator.eq,
        ast.NotEq: operator.ne,
        ast.Lt: operator.lt,
        ast.LtE: operator.le,
        ast.Gt: operator.gt,
        ast.GtE: operator.ge,
        ast.In: lambda x, y: y.isin(x),
        ast.NotIn: lambda x, y: ~y.isin(x),
        ast.UAdd: operator.pos,
        ast.USub: operator.neg,
        ast.Invert: operator.invert,
        ast.And: operator.and_,
        ast.Or: operator.or_,
    }

    def __init__(self, resolvers, target, env):
        self.env = env
        self.target = target
        self.resolvers = resolvers

        self.referenced_vars = set()
        self.assigned = False
        self.entity_subscribe = False

    def _preparse(self, expr):
        reader = StringIO(expr).readline
        return tokenize.untokenize(list(_tokenize_str(reader)))

    def eval(self, expr, rewrite=True):
        if rewrite:
            expr = self._preparse(expr)
        node = ast.fix_missing_locations(ast.parse(expr))
        return self.visit(node)

    def get_named_object(self, obj_name):
        for resolver in self.resolvers:
            try:
                return resolver[obj_name]
            except (IndexError, KeyError):
                continue
        if obj_name in self.env:
            self.referenced_vars.add(obj_name)
            return self.env[obj_name]
        raise KeyError(f"name {obj_name} is not defined")

    def visit(self, node):
        if isinstance(node, ENTITY_TYPE):
            return node
        node_name = node.__class__.__name__
        method = "visit_" + node_name
        try:
            visitor = getattr(self, method)
        except AttributeError:
            raise SyntaxError(
                "Query string contains unsupported syntax: {}".format(node_name)
            )
        return visitor(node)

    def visit_Module(self, node):
        if self.target is None and len(node.body) != 1:
            raise SyntaxError("Only a single expression is allowed")
        result = None
        for expr in node.body:
            result = self.visit(expr)
        return result

    def visit_Expr(self, node):
        return self.visit(node.value)

    def visit_BinOp(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return self._op_handlers[type(node.op)](left, right)

    def visit_Call(self, node):
        func = self.visit(node.func)
        args = [self.visit(n) for n in node.args]
        kwargs = OrderedDict([(kw.arg, self.visit(kw.value)) for kw in node.keywords])
        return func(*args, **kwargs)

    def visit_Compare(self, node):
        ops = node.ops
        comps = node.comparators

        if len(comps) == 1:
            binop = ast.BinOp(op=ops[0], left=node.left, right=comps[0])
            return self.visit(binop)

        left = node.left
        values = []
        for op, comp in zip(ops, comps):
            new_node = ast.Compare(comparators=[comp], left=left, ops=[op])
            left = comp
            values.append(new_node)
        return self.visit(ast.BoolOp(op=ast.And(), values=values))

    def visit_BoolOp(self, node):
        def func(lhs, rhs):
            binop = ast.BinOp(op=node.op, left=lhs, right=rhs)
            return self.visit(binop)

        return reduce(func, node.values)

    def visit_UnaryOp(self, node):
        op = self.visit(node.operand)
        return self._op_handlers[type(node.op)](op)

    def visit_Name(self, node):
        if node.id.startswith(LOCAL_TAG):
            local_name = node.id.replace(LOCAL_TAG, "")
            self.referenced_vars.add(local_name)
            return self.env[local_name]
        if node.id.startswith(BACKTICK_TAG):
            local_name = binascii.a2b_hex(
                node.id.replace(BACKTICK_TAG, "").encode()
            ).decode()
            return self.get_named_object(local_name)
        return self.get_named_object(node.id)

    def visit_NameConstant(self, node):  # pragma: no cover
        return node.value

    def visit_Num(self, node):  # pragma: no cover
        return node.n

    def visit_Str(self, node):  # pragma: no cover
        return node.s

    def visit_Constant(self, node):
        return node.value

    def visit_List(self, node):
        return [self.visit(e) for e in node.elts]

    def visit_Assign(self, node):
        if self.target is None:
            raise ValueError("Target not specified for assignment")
        if isinstance(node.targets[0], ast.Tuple):
            raise ValueError("Does not support assigning to multiple objects")

        target = node.targets[0].id
        value = self.visit(node.value)
        self.target[target] = value
        self.assigned = True

    visit_Tuple = visit_List

    def visit_Attribute(self, node):
        attr = node.attr
        value = node.value

        ctx = node.ctx
        if isinstance(ctx, ast.Load):
            resolved = self.visit(value)
            return getattr(resolved, attr)

        raise ValueError("Invalid Attribute context {0}".format(ctx.__name__))

    def visit_Subscript(self, node):
        value = self.visit(node.value)
        sub = self.visit(node.slice)
        if isinstance(value, ENTITY_TYPE):
            self.entity_subscribe = True
        return value[sub]

    def visit_Index(self, node):
        return self.visit(node.value)

    def visit_Slice(self, node):
        lower = node.lower
        if lower is not None:
            lower = self.visit(lower)
        upper = node.upper
        if upper is not None:
            upper = self.visit(upper)
        step = node.step
        if step is not None:
            step = self.visit(step)

        return slice(lower, upper, step)


class DataFrameEval(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.DATAFRAME_EVAL

    _expr = StringField("expr")
    _parser = StringField("parser")
    _engine = StringField("engine")
    _variables = DictField("variables")
    _self_target = BoolField("self_target")
    _is_query = BoolField("is_query")

    def __init__(
        self,
        expr=None,
        parser=None,
        engine=None,
        variables=None,
        self_target=None,
        is_query=None,
        **kw,
    ):
        super().__init__(
            _expr=expr,
            _parser=parser,
            _engine=engine,
            _variables=variables,
            _self_target=self_target,
            _is_query=is_query,
            **kw,
        )

    @property
    def expr(self):
        return self._expr

    @property
    def parser(self):
        return self._parser

    @property
    def engine(self):
        return self._engine

    @property
    def variables(self):
        return self._variables

    @property
    def self_target(self):
        return self._self_target

    @property
    def is_query(self):
        return self._is_query

    def __call__(self, df, output_type, shape, dtypes):
        self._output_types = [output_type]
        params = df.params
        new_index_value = (
            df.index_value if not np.isnan(shape[0]) else parse_index(pd.RangeIndex(-1))
        )
        if output_type == OutputType.dataframe:
            params.update(
                dict(
                    dtypes=dtypes,
                    shape=shape,
                    columns_value=parse_index(dtypes.index, store_data=True),
                    index_value=new_index_value,
                )
            )
        else:
            name, dtype = dtypes
            params = dict(
                name=name,
                dtype=dtype,
                shape=shape,
                index_value=new_index_value,
            )
        return self.new_tileable([df], **params)

    def convert_to_query(self, df, output_type, shape, dtypes):
        new_op = self.copy().reset_key()
        new_op._is_query = True
        new_op._self_target = False
        return new_op(df, output_type, shape, dtypes)

    @classmethod
    def tile(cls, op: "DataFrameEval"):
        in_df = op.inputs[0]
        out_df = op.outputs[0]

        if in_df.ndim == 2:
            if in_df.chunk_shape[1] > 1:
                in_df = yield from recursive_tile(in_df.rechunk({1: in_df.shape[1]}))

        chunks = []
        for c in in_df.chunks:
            if out_df.ndim == 2:
                new_shape = (
                    np.nan if np.isnan(out_df.shape[0]) else c.shape[0],
                    out_df.shape[1],
                )
                params = dict(
                    dtypes=out_df.dtypes,
                    shape=new_shape,
                    columns_value=parse_index(out_df.dtypes.index, store_data=True),
                    index_value=c.index_value,
                    index=c.index,
                )
            else:
                new_shape = (np.nan if np.isnan(out_df.shape[0]) else c.shape[0],)
                params = dict(
                    name=out_df.name,
                    dtype=out_df.dtype,
                    shape=new_shape,
                    index_value=c.index_value,
                    index=(c.index[0],),
                )
            new_op = op.copy().reset_key()
            chunks.append(new_op.new_chunk([c], **params))

        new_op = op.copy().reset_key()
        params = out_df.params

        new_nsplits = [in_df.nsplits[0], (out_df.shape[-1],)]
        if np.isnan(out_df.shape[0]):
            new_nsplits[0] = (np.nan,) * len(in_df.nsplits[0])
        if out_df.ndim == 1:
            new_nsplits = new_nsplits[:1]

        params.update(
            dict(
                chunks=chunks,
                nsplits=tuple(new_nsplits),
            )
        )
        return new_op.new_tileables([in_df], **params)

    @classmethod
    def execute(cls, ctx, op: "DataFrameEval"):
        in_data = ctx[op.inputs[0].key]

        if op.self_target:
            in_data = in_data.copy()

        if op.is_query:
            val = in_data.query(
                op.expr, parser=op.parser, engine=op.engine, local_dict=op.variables
            )
        else:
            val = in_data.eval(
                op.expr, parser=op.parser, engine=op.engine, local_dict=op.variables
            )
        ctx[op.outputs[0].key] = val


def mars_eval(
    expr,
    parser="mars",
    engine=None,
    local_dict=None,
    global_dict=None,
    resolvers=(),
    level=0,
    target=None,
    inplace=False,
):
    """

    Evaluate a Python expression as a string using various backends.

    The following arithmetic operations are supported: ``+``, ``-``, ``*``,
    ``/``, ``**``, ``%``, ``//`` (python engine only) along with the following
    boolean operations: ``|`` (or), ``&`` (and), and ``~`` (not).
    Additionally, the ``'pandas'`` parser allows the use of :keyword:`and`,
    :keyword:`or`, and :keyword:`not` with the same semantics as the
    corresponding bitwise operators.  :class:`~pandas.Series` and
    :class:`~pandas.DataFrame` objects are supported and behave as they would
    with plain ol' Python evaluation.

    Parameters
    ----------
    expr : str
        The expression to evaluate. This string cannot contain any Python
        `statements
        <https://docs.python.org/3/reference/simple_stmts.html#simple-statements>`__,
        only Python `expressions
        <https://docs.python.org/3/reference/simple_stmts.html#expression-statements>`__.
    local_dict : dict or None, optional
        A dictionary of local variables, taken from locals() by default.
    global_dict : dict or None, optional
        A dictionary of global variables, taken from globals() by default.
    resolvers : list of dict-like or None, optional
        A list of objects implementing the ``__getitem__`` special method that
        you can use to inject an additional collection of namespaces to use for
        variable lookup. For example, this is used in the
        :meth:`~DataFrame.query` method to inject the
        ``DataFrame.index`` and ``DataFrame.columns``
        variables that refer to their respective :class:`~pandas.DataFrame`
        instance attributes.
    level : int, optional
        The number of prior stack frames to traverse and add to the current
        scope. Most users will **not** need to change this parameter.
    target : object, optional, default None
        This is the target object for assignment. It is used when there is
        variable assignment in the expression. If so, then `target` must
        support item assignment with string keys, and if a copy is being
        returned, it must also support `.copy()`.
    inplace : bool, default False
        If `target` is provided, and the expression mutates `target`, whether
        to modify `target` inplace. Otherwise, return a copy of `target` with
        the mutation.

    Returns
    -------
    ndarray, numeric scalar, DataFrame, Series

    Raises
    ------
    ValueError
        There are many instances where such an error can be raised:

        - `target=None`, but the expression is multiline.
        - The expression is multiline, but not all them have item assignment.
          An example of such an arrangement is this:

          a = b + 1
          a + 2

          Here, there are expressions on different lines, making it multiline,
          but the last line has no variable assigned to the output of `a + 2`.
        - `inplace=True`, but the expression is missing item assignment.
        - Item assignment is provided, but the `target` does not support
          string item assignment.
        - Item assignment is provided and `inplace=False`, but the `target`
          does not support the `.copy()` method

    See Also
    --------
    DataFrame.query : Evaluates a boolean expression to query the columns
            of a frame.
    DataFrame.eval : Evaluate a string describing operations on
            DataFrame columns.

    Notes
    -----
    The ``dtype`` of any objects involved in an arithmetic ``%`` operation are
    recursively cast to ``float64``.

    See the :ref:`enhancing performance <enhancingperf.eval>` documentation for
    more details.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> df = md.DataFrame({"animal": ["dog", "pig"], "age": [10, 20]})
    >>> df.execute()
      animal  age
    0    dog   10
    1    pig   20

    We can add a new column using ``pd.eval``:

    >>> md.eval("double_age = df.age * 2", target=df).execute()
      animal  age  double_age
    0    dog   10          20
    1    pig   20          40
    """
    if not isinstance(expr, str):
        raise TypeError("expr must be a string")

    expr = textwrap.dedent(expr)

    try:
        frame = sys._getframe(level + 1)
        local_dict = local_dict or dict()
        local_dict.update(frame.f_locals)
        global_dict = global_dict or dict()
        global_dict.update(frame.f_globals)
    finally:
        del frame

    env = dict()
    env.update(global_dict)
    env.update(local_dict)

    ref_frames = set(resolvers) | set([target] if target is not None else [])
    self_target = len(resolvers) > 0 and resolvers[0] is target

    if target is not None and not inplace:
        target = target.copy()

    visitor = CollectionVisitor(resolvers, target, env)
    result = visitor.eval(expr)
    result = result if result is not None else target
    has_var_frame = any(
        isinstance(env[k], ENTITY_TYPE) for k in visitor.referenced_vars
    )
    if len(ref_frames) != 1 or visitor.entity_subscribe or has_var_frame:
        if parser != "mars":
            raise NotImplementedError("Does not support parser names other than mars")
        if engine is not None:
            raise NotImplementedError("Does not support specifying engine names")
        return result
    else:
        parser = "pandas" if parser == "mars" else parser
        referenced_env = {k: env[k] for k in visitor.referenced_vars}
        op = DataFrameEval(
            expr,
            parser=parser,
            engine=engine,
            variables=referenced_env,
            self_target=visitor.assigned and self_target,
            is_query=False,
        )
        output_type = get_output_types(result)[0]
        dtypes = result.dtypes if result.ndim == 2 else (result.name, result.dtype)
        return op(resolvers[0], output_type, result.shape, dtypes)


def df_eval(df, expr, inplace=False, **kwargs):
    """
    Evaluate a string describing operations on DataFrame columns.

    Operates on columns only, not specific rows or elements.  This allows
    `eval` to run arbitrary code, which can make you vulnerable to code
    injection if you pass user input to this function.

    Parameters
    ----------
    expr : str
        The expression string to evaluate.
    inplace : bool, default False
        If the expression contains an assignment, whether to perform the
        operation inplace and mutate the existing DataFrame. Otherwise,
        a new DataFrame is returned.
    **kwargs
        See the documentation for :func:`eval` for complete details
        on the keyword arguments accepted by
        :meth:`~pandas.DataFrame.query`.

    Returns
    -------
    ndarray, scalar, or pandas object
        The result of the evaluation.

    See Also
    --------
    DataFrame.query : Evaluates a boolean expression to query the columns
        of a frame.
    DataFrame.assign : Can evaluate an expression or function to create new
        values for a column.
    eval : Evaluate a Python expression as a string using various
        backends.

    Notes
    -----
    For more details see the API documentation for :func:`~eval`.
    For detailed examples see :ref:`enhancing performance with eval
    <enhancingperf.eval>`.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> df = md.DataFrame({'A': range(1, 6), 'B': range(10, 0, -2)})
    >>> df.execute()
       A   B
    0  1  10
    1  2   8
    2  3   6
    3  4   4
    4  5   2
    >>> df.eval('A + B').execute()
    0    11
    1    10
    2     9
    3     8
    4     7
    dtype: int64

    Assignment is allowed though by default the original DataFrame is not
    modified.

    >>> df.eval('C = A + B').execute()
       A   B   C
    0  1  10  11
    1  2   8  10
    2  3   6   9
    3  4   4   8
    4  5   2   7
    >>> df.execute()
       A   B
    0  1  10
    1  2   8
    2  3   6
    3  4   4
    4  5   2

    Use ``inplace=True`` to modify the original DataFrame.

    >>> df.eval('C = A + B', inplace=True)
    >>> df.execute()
       A   B   C
    0  1  10  11
    1  2   8  10
    2  3   6   9
    3  4   4   8
    4  5   2   7

    Multiple columns can be assigned to using multi-line expressions:

    >>> df.eval('''
    ... C = A + B
    ... D = A - B
    ... ''').execute()
       A   B   C  D
    0  1  10  11 -9
    1  2   8  10 -6
    2  3   6   9 -3
    3  4   4   8  0
    4  5   2   7  3
    """
    level = kwargs.pop("level", None) or 0
    kwargs["inplace"] = inplace
    val = mars_eval(expr, resolvers=(df,), target=df, level=level + 1, **kwargs)
    if not inplace:
        return val


def df_query(df, expr, inplace=False, **kwargs):
    """
    Query the columns of a DataFrame with a boolean expression.

    Parameters
    ----------
    expr : str
        The query string to evaluate.

        You can refer to variables
        in the environment by prefixing them with an '@' character like
        ``@a + b``.

        You can refer to column names that contain spaces or operators by
        surrounding them in backticks. This way you can also escape
        names that start with a digit, or those that  are a Python keyword.
        Basically when it is not valid Python identifier. See notes down
        for more details.

        For example, if one of your columns is called ``a a`` and you want
        to sum it with ``b``, your query should be ```a a` + b``.

    inplace : bool
        Whether the query should modify the data in place or return
        a modified copy.
    **kwargs
        See the documentation for :func:`eval` for complete details
        on the keyword arguments accepted by :meth:`DataFrame.query`.

    Returns
    -------
    DataFrame
        DataFrame resulting from the provided query expression.

    See Also
    --------
    eval : Evaluate a string describing operations on
        DataFrame columns.
    DataFrame.eval : Evaluate a string describing operations on
        DataFrame columns.

    Notes
    -----
    The result of the evaluation of this expression is first passed to
    :attr:`DataFrame.loc` and if that fails because of a
    multidimensional key (e.g., a DataFrame) then the result will be passed
    to :meth:`DataFrame.__getitem__`.

    This method uses the top-level :func:`eval` function to
    evaluate the passed query.

    The :meth:`~pandas.DataFrame.query` method uses a slightly
    modified Python syntax by default. For example, the ``&`` and ``|``
    (bitwise) operators have the precedence of their boolean cousins,
    :keyword:`and` and :keyword:`or`. This *is* syntactically valid Python,
    however the semantics are different.

    You can change the semantics of the expression by passing the keyword
    argument ``parser='python'``. This enforces the same semantics as
    evaluation in Python space. Likewise, you can pass ``engine='python'``
    to evaluate an expression using Python itself as a backend. This is not
    recommended as it is inefficient compared to using ``numexpr`` as the
    engine.

    The :attr:`DataFrame.index` and
    :attr:`DataFrame.columns` attributes of the
    :class:`~pandas.DataFrame` instance are placed in the query namespace
    by default, which allows you to treat both the index and columns of the
    frame as a column in the frame.
    The identifier ``index`` is used for the frame index; you can also
    use the name of the index to identify it in a query. Please note that
    Python keywords may not be used as identifiers.

    For further details and examples see the ``query`` documentation in
    :ref:`indexing <indexing.query>`.

    *Backtick quoted variables*

    Backtick quoted variables are parsed as literal Python code and
    are converted internally to a Python valid identifier.
    This can lead to the following problems.

    During parsing a number of disallowed characters inside the backtick
    quoted string are replaced by strings that are allowed as a Python identifier.
    These characters include all operators in Python, the space character, the
    question mark, the exclamation mark, the dollar sign, and the euro sign.
    For other characters that fall outside the ASCII range (U+0001..U+007F)
    and those that are not further specified in PEP 3131,
    the query parser will raise an error.
    This excludes whitespace different than the space character,
    but also the hashtag (as it is used for comments) and the backtick
    itself (backtick can also not be escaped).

    In a special case, quotes that make a pair around a backtick can
    confuse the parser.
    For example, ```it's` > `that's``` will raise an error,
    as it forms a quoted string (``'s > `that'``) with a backtick inside.

    See also the Python documentation about lexical analysis
    (https://docs.python.org/3/reference/lexical_analysis.html)
    in combination with the source code in :mod:`pandas.core.computation.parsing`.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> df = md.DataFrame({'A': range(1, 6),
    ...                    'B': range(10, 0, -2),
    ...                    'C C': range(10, 5, -1)})
    >>> df.execute()
       A   B  C C
    0  1  10   10
    1  2   8    9
    2  3   6    8
    3  4   4    7
    4  5   2    6
    >>> df.query('A > B').execute()
       A  B  C C
    4  5  2    6

    The previous expression is equivalent to

    >>> df[df.A > df.B].execute()
       A  B  C C
    4  5  2    6

    For columns with spaces in their name, you can use backtick quoting.

    >>> df.query('B == `C C`').execute()
       A   B  C C
    0  1  10   10

    The previous expression is equivalent to

    >>> df[df.B == df['C C']].execute()
       A   B  C C
    0  1  10   10
    """
    level = kwargs.pop("level", None) or 0
    predicate = mars_eval(expr, resolvers=(df,), level=level + 1, **kwargs)
    result = df[predicate]

    if isinstance(predicate.op, DataFrameEval):
        output_type = get_output_types(result)[0]
        dtypes = result.dtypes if result.ndim == 2 else (result.name, result.dtype)
        result = predicate.op.convert_to_query(df, output_type, result.shape, dtypes)

    if inplace:
        df.data = result.data
    else:
        return result



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Any, List

from ....storage.base import StorageLevel
from ..core import DataInfo


class AbstractStorageAPI(ABC):
    @abstractmethod
    async def get(
        self, data_key: str, conditions: List = None, error: str = "raise"
    ) -> Any:
        """
        Get object by data key.

        Parameters
        ----------
        data_key: str
            date key to get.

        conditions: List
            Index conditions to pushdown

        error: str
            raise or ignore

        Returns
        -------
            object
        """

    @abstractmethod
    async def put(
        self, data_key: str, obj: object, level: StorageLevel = StorageLevel.MEMORY
    ) -> DataInfo:
        """
        Put object into storage.

        Parameters
        ----------
        data_key: str
            data key to put.
        obj: object
            object to put.
        level: StorageLevel
            the storage level to put into, MEMORY as default

        Returns
        -------
        object information: ObjectInfo
            the put object information
        """

    @abstractmethod
    async def get_infos(self, data_key: str) -> List[DataInfo]:
        """
        Get data information items for specific data key

        Parameters
        ----------
        data_key

        Returns
        -------
        out
            List of information for specified key
        """



# -*- coding: utf-8 -*-

from collections import Counter
from hashlib import md5


class MetaRing(object):
    """Implement a tunable consistent hashing ring."""

    def __init__(self, hash_fn):
        """Create a new HashRing.

        :param hash_fn: use this callable function to hash keys.
        """
        self._distribution = Counter()
        self._keys = []
        self._nodes = {}
        self._ring = {}

        if hash_fn and not hasattr(hash_fn, "__call__"):
            raise TypeError("hash_fn should be a callable function")
        self._hash_fn = hash_fn or (
            lambda key: int(md5(str(key).encode("utf-8")).hexdigest(), 16)
        )

    def hashi(self, key):
        """Returns an integer derived from the md5 hash of the given key."""
        return self._hash_fn(key)

    def _create_ring(self, nodes):
        """Generate a ketama compatible continuum/ring."""
        for node_name, node_conf in nodes:
            for w in range(0, node_conf["vnodes"] * node_conf["weight"]):
                self._distribution[node_name] += 1
                self._ring[self.hashi(f"{node_name}-{w}")] = node_name
        self._keys = sorted(self._ring.keys())

    def _remove_node(self, node_name):
        """Remove the given node from the continuum/ring.

        :param node_name: the node name.
        """
        try:
            node_conf = self._nodes.pop(node_name)
        except Exception:
            raise KeyError(
                f"node '{node_name}' not found, "
                f"available nodes: {list(self._nodes.keys())}"
            )
        else:
            self._distribution.pop(node_name)
            for w in range(0, node_conf["vnodes"] * node_conf["weight"]):
                del self._ring[self.hashi(f"{node_name}-{w}")]
            self._keys = sorted(self._ring.keys())



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import List, Tuple

from ... import oscar as mo
from ...storage import StorageLevel
from .core import DataManagerActor
from .errors import NoDataToSpill
from .handler import StorageHandlerActor

logger = logging.getLogger(__name__)

DEFAULT_SPILL_BLOCK_SIZE = 128 * 1024


class SpillStrategy(ABC):
    @abstractmethod
    def record_put_info(self, key, data_size: int):
        """
        Record the data key and data size when putting into storage
        """

    @abstractmethod
    def record_delete_info(self, key):
        """
        Record who is removed from storage
        """

    @abstractmethod
    def get_spill_keys(self, size: int) -> Tuple[List, List]:
        """
        Return sizes and keys for spilling according to spill size
        """


class FIFOStrategy(SpillStrategy):
    def __init__(self, level: StorageLevel):
        self._level = level
        self._data_sizes = dict()
        self._pinned_keys = defaultdict(int)
        self._spilling_keys = set()

    def record_put_info(self, key, data_size: int):
        self._data_sizes[key] = data_size

    def record_delete_info(self, key):
        self._data_sizes.pop(key, None)
        if key in self._spilling_keys:
            self._spilling_keys.remove(key)

    def pin_data(self, key):
        self._pinned_keys[key] += 1

    def unpin_data(self, key):
        if key not in self._pinned_keys:
            return
        self._pinned_keys[key] -= 1
        if self._pinned_keys[key] <= 0:
            del self._pinned_keys[key]

    def get_spillable_size(self):
        total_size = 0
        for data_key, data_size in self._data_sizes.items():
            if (
                data_key not in self._pinned_keys
                and data_key not in self._spilling_keys
            ):
                total_size += data_size
        return total_size

    def get_spill_keys(self, size: int) -> Tuple[List, List]:
        spill_sizes = []
        spill_keys = []
        spill_size = 0
        for data_key, data_size in self._data_sizes.items():
            if spill_size >= size:
                break
            if data_key in self._pinned_keys:
                continue
            if data_key in self._spilling_keys:
                continue
            spill_sizes.append(data_size)
            spill_keys.append(data_key)
            spill_size += data_size

        if spill_size < size:  # pragma: no cover
            pinned_sizes = dict((k, self._data_sizes[k]) for k in self._pinned_keys)
            spilling_keys = dict((k, self._data_sizes[k]) for k in self._spilling_keys)
            logger.debug(
                "No data can be spilled for level: %s, pinned keys: %s,"
                " spilling keys: %s",
                self._level,
                pinned_sizes,
                spilling_keys,
            )
            raise NoDataToSpill(f"No data can be spilled for level: {self._level}")
        self._spilling_keys.update(set(spill_keys))
        return spill_sizes, spill_keys


class SpillManagerActor(mo.StatelessActor):
    """
    The actor to handle the race condition when NoDataToSpill happens.
    There are two situations when spill raises `NoDataToSpill`,
    one is that space is allocated while objects are not put into storage,
    another is some objects are pinned that can not be spilled,
    so we create an asyncio event if not have enough objects to spill,
    when put or unpin happens, we will notify and check spillable size,
    if size is enough for spilling, call event.set() to wake up spilling task.
    """

    def __init__(self, level: StorageLevel):
        self._level = level
        self._event = None
        self._lock = asyncio.Lock()

    @classmethod
    def gen_uid(cls, band_name: str, level: StorageLevel):
        return f"spill_manager_{band_name}_{level}"

    def has_spill_task(self):
        return self._event is not None

    def notify_spillable_space(self, spillable_size: int, quota_left: int):
        event = self._event
        if event is None:
            return
        logger.debug("Notify to check if has space for spilling")
        if spillable_size + quota_left > event.size:
            logger.debug(
                "Check pass, wake up spill task, spill bytes is %s",
                event.size - quota_left,
            )
            event.size = event.size - quota_left
            event.set()

    async def wait_for_space(self, size: int):
        # make sure only one spilling task is waiting the event
        async with self._lock:
            self._event = event = asyncio.Event()
            event.size = size
            await self._event.wait()
            size = self._event.size
            self._event = None
            return size


async def spill(
    request_size: int,
    level: StorageLevel,
    band_name: str,
    data_manager: mo.ActorRefType[DataManagerActor],
    storage_handler: mo.ActorRefType[StorageHandlerActor],
    block_size=None,
    multiplier=1.1,
):
    logger.debug(
        "%s is full, need to spill %s bytes, multiplier is %s",
        level,
        request_size,
        multiplier,
    )
    request_size *= multiplier
    block_size = block_size or DEFAULT_SPILL_BLOCK_SIZE
    spill_level = level.spill_level()
    spill_sizes, spill_keys = await data_manager.get_spill_keys(
        level, band_name, request_size
    )
    logger.debug(
        "Decide to spill %s bytes, data keys are %s", sum(spill_sizes), spill_keys
    )

    for (session_id, key), size in zip(spill_keys, spill_sizes):
        reader = await storage_handler.open_reader(session_id, key)
        writer = await storage_handler.open_writer(session_id, key, size, spill_level)
        async with reader:
            async with writer:
                while True:
                    block_data = await reader.read(block_size)
                    if not block_data:
                        break
                    else:
                        await writer.write(block_data)
        try:
            await storage_handler.delete_object(
                session_id, key, size, reader.object_id, level
            )
        except KeyError:  # pragma: no cover
            # workaround for the case that the object
            # has been deleted during spill
            logger.debug("Data %s %s is deleted during spill", session_id, key)
            await storage_handler.delete(session_id, key, error="ignore")
    logger.debug("Spill finishes, release %s bytes of %s", sum(spill_sizes), level)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import operator
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Tuple, Type, Union

from ..utils import dataslots
from .core import StorageFileObject

_storage_backends = dict()


def register_storage_backend(backend: Type["StorageBackend"]):
    _storage_backends[backend.name] = backend
    return backend


def get_storage_backend(backend_name) -> Type["StorageBackend"]:
    return _storage_backends[backend_name]


_ComparableLevel = Union[int, "StorageLevel"]


class StorageLevel(Enum):
    GPU = 1 << 0
    MEMORY = 1 << 1
    DISK = 1 << 2
    REMOTE = 1 << 3

    def __and__(self, other: _ComparableLevel):
        other_value = getattr(other, "value", other)
        return self.value & other_value

    __rand__ = __and__

    def __or__(self, other: _ComparableLevel):
        other_value = getattr(other, "value", other)
        return self.value | other_value

    __ror__ = __or__

    def __lt__(self, other: _ComparableLevel):
        other_value = getattr(other, "value", other)
        return self.value < other_value

    def __gt__(self, other: _ComparableLevel):
        other_value = getattr(other, "value", other)
        return self.value > other_value

    def spill_level(self):
        if self == StorageLevel.GPU:
            return StorageLevel.MEMORY
        elif self == StorageLevel.MEMORY:
            return StorageLevel.DISK
        else:  # pragma: no cover
            raise ValueError(f"Level {self} doesn't have spill level")

    @staticmethod
    def from_str(s: str):
        level_mapping = StorageLevel.__members__
        level_strings = [ss.strip() for ss in s.upper().split("|")]
        levels = []
        for ls in level_strings:
            if ls not in level_mapping:  # pragma: no cover
                raise ValueError(f"Unknown level {ls}")
            levels.append(level_mapping[ls])
        return functools.reduce(operator.or_, levels)


@dataslots
@dataclass
class ObjectInfo:
    size: int = None
    device: int = None
    object_id: Any = None


class StorageBackend(ABC):
    name = None

    @classmethod
    @abstractmethod
    async def setup(cls, **kwargs) -> Tuple[Dict, Dict]:
        """
        Setup environments, for example, start plasma store for plasma backend.

        Parameters
        ----------
        kwargs : kwargs
            Kwargs for setup.

        Returns
        -------
        Tuple of two dicts
            Dicts for initialization and teardown.
        """

    @staticmethod
    async def teardown(**kwargs):
        """
        Clean up the environments.

        Parameters
        ----------
        kwargs : kwargs
             Parameters for clean up.
        """

    @property
    def size(self) -> Union[int, None]:
        """
        The total size of storage.

        Returns
        -------
        Size: int
            Total size of storage.
        """
        return None

    @property
    @abstractmethod
    def level(self) -> StorageLevel:
        """
        Level of current storage backend.

        Returns
        -------
        Level: StorageLevel
            storage level.
        """

    @property
    def backend_info(self) -> dict:
        """
        Get the customized backend info of this storage backend.

        Returns
        -------
        info: dict
            Customized storage backend info dict.
        """
        return {"name": self.name}

    @abstractmethod
    async def get(self, object_id, **kwargs) -> object:
        """
        Get object by key. For some backends, `columns` or `slice` can pass to get part of data.

        Parameters
        ----------
        object_id : object id
            Object id to get.

        kwargs:
            Additional keyword arguments

        Returns
        -------
        Python object
        """

    @abstractmethod
    async def put(self, obj, importance: int = 0) -> ObjectInfo:
        """
        Put object into storage with object_id.

        Parameters
        ----------
        obj : python object
            Object to put.

        importance: int
             The priority to spill when storage is full

        Returns
        -------
        ObjectInfo
            object information including size, raw_size, device
        """

    @abstractmethod
    async def delete(self, object_id):
        """
        Delete object from storage by object_id.

        Parameters
        ----------
        object_id
            object id
        """

    @abstractmethod
    async def object_info(self, object_id) -> ObjectInfo:
        """
        Get information about stored object.

        Parameters
        ----------
        object_id
            object id

        Returns
        -------
        ObjectInfo
            Object info including size, device and etc.
        """

    @abstractmethod
    async def open_writer(self, size=None) -> StorageFileObject:
        """
        Return a file-like object for writing.

        Parameters
        ----------
        size: int
            Maximum size in bytes

        Returns
        -------
        fileobj: StorageFileObject
        """

    @abstractmethod
    async def open_reader(self, object_id) -> StorageFileObject:
        """
        Return a file-like object for reading.

        Parameters
        ----------
        object_id
            Object id

        Returns
        -------
        fileobj: StorageFileObject
        """

    async def list(self) -> List:
        """
        List all stored objects in storage.

        Returns
        -------
        List of objects
        """

    async def fetch(self, object_id):
        """
        Fetch object to current worker.

        Parameters
        ----------
        object_id
            Object id.
        """

    async def pin(self, object_id):
        """
        Pin the data to prevent the data being released or spilled.

        Parameters
        ----------
        object_id
            object id
        """

    async def unpin(self, object_id):
        """
        Unpin the data, allow storage to release the data.

        Parameters
        ----------
        object_id
            object id
        """



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import inject_dtype
from .core import TensorUnaryOp
from .utils import arithmetic_operand


@arithmetic_operand(sparse_mode="unary")
class TensorIsInf(TensorUnaryOp):
    _op_type_ = OperandDef.ISINF
    _func_name = "isinf"


@inject_dtype(np.bool_)
def isinf(x, out=None, where=None, **kwargs):
    """
    Test element-wise for positive or negative infinity.

    Returns a boolean array of the same shape as `x`, True where ``x ==
    +/-inf``, otherwise False.

    Parameters
    ----------
    x : array_like
        Input values
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    y : bool (scalar) or boolean Tensor
        For scalar input, the result is a new boolean with value True if
        the input is positive or negative infinity; otherwise the value is
        False.

        For tensor input, the result is a boolean tensor with the same shape
        as the input and the values are True where the corresponding
        element of the input is positive or negative infinity; elsewhere
        the values are False.  If a second argument was supplied the result
        is stored there.  If the type of that array is a numeric type the
        result is represented as zeros and ones, if the type is boolean
        then as False and True, respectively.  The return value `y` is then
        a reference to that tensor.

    See Also
    --------
    isneginf, isposinf, isnan, isfinite

    Notes
    -----
    Mars uses the IEEE Standard for Binary Floating-Point for Arithmetic
    (IEEE 754).

    Errors result if the second argument is supplied when the first
    argument is a scalar, or if the first and second arguments have
    different shapes.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.isinf(mt.inf).execute()
    True
    >>> mt.isinf(mt.nan).execute()
    False
    >>> mt.isinf(mt.NINF).execute()
    True
    >>> mt.isinf([mt.inf, -mt.inf, 1.0, mt.nan]).execute()
    array([ True,  True, False, False])

    >>> x = mt.array([-mt.inf, 0., mt.inf])
    >>> y = mt.array([2, 2, 2])
    >>> mt.isinf(x, y).execute()
    array([1, 0, 1])
    >>> y.execute()
    array([1, 0, 1])
    """
    op = TensorIsInf(**kwargs)
    return op(x, out=out, where=where)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect
import threading

from ..config import options

_internal_mode = threading.local()


def is_eager_mode():
    in_kernel = is_kernel_mode()
    if not in_kernel:
        return options.eager_mode
    else:
        # in kernel, eager mode always False
        return False


def is_kernel_mode():
    try:
        return bool(_internal_mode.kernel)
    except AttributeError:
        _internal_mode.kernel = None
        return False


def is_build_mode():
    return bool(getattr(_internal_mode, "build", False))


class _EnterModeFuncWrapper:
    def __init__(self, mode_name_to_value):
        self.mode_name_to_value = mode_name_to_value

        # as the wrapper may enter for many times
        # record old values for each time
        self.mode_name_to_value_list = list()

    def __enter__(self):
        mode_name_to_old_value = dict()
        for mode_name, value in self.mode_name_to_value.items():
            # record mode's old values
            mode_name_to_old_value[mode_name] = getattr(_internal_mode, mode_name, None)
            if value is None:
                continue
            # set value
            setattr(_internal_mode, mode_name, value)
        self.mode_name_to_value_list.append(mode_name_to_old_value)

    def __exit__(self, *_):
        mode_name_to_old_value = self.mode_name_to_value_list.pop()
        for mode_name in self.mode_name_to_value.keys():
            # set back old values
            setattr(_internal_mode, mode_name, mode_name_to_old_value[mode_name])

    def __call__(self, func):
        mode_name_to_value = self.mode_name_to_value.copy()
        if not inspect.iscoroutinefunction(func):
            # sync
            @functools.wraps(func)
            def _inner(*args, **kwargs):
                with enter_mode(**mode_name_to_value):
                    return func(*args, **kwargs)

        else:
            # async
            @functools.wraps(func)
            async def _inner(*args, **kwargs):
                with enter_mode(**mode_name_to_value):
                    return await func(*args, **kwargs)

        return _inner


def enter_mode(kernel=None, build=None):
    mode_name_to_value = {
        "kernel": kernel,
        "build": build,
    }
    mode_name_to_value = {k: v for k, v in mode_name_to_value.items() if v is not None}

    return _EnterModeFuncWrapper(mode_name_to_value)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
from typing import List, Dict, Optional, Set, Type, TypeVar

from .... import oscar as mo
from ....lib.aio import alru_cache
from ....resource import Resource
from ....typing import BandType
from ...core import NodeRole
from ..core import (
    watch_method,
    NodeStatus,
    WorkerSlotInfo,
    QuotaInfo,
    DiskInfo,
    StorageInfo,
)
from .core import AbstractClusterAPI

APIType = TypeVar("APIType", bound="ClusterAPI")
logger = logging.getLogger(__name__)


class ClusterAPI(AbstractClusterAPI):
    def __init__(self, address: str):
        self._address = address
        self._locator_ref = None
        self._uploader_ref = None

    async def _init(self):
        from ..locator import SupervisorLocatorActor
        from ..uploader import NodeInfoUploaderActor

        self._locator_ref = await mo.actor_ref(
            SupervisorLocatorActor.default_uid(), address=self._address
        )
        self._uploader_ref = await mo.actor_ref(
            NodeInfoUploaderActor.default_uid(), address=self._address
        )

    @classmethod
    @alru_cache(cache_exceptions=False)
    async def create(cls: Type[APIType], address: str) -> APIType:
        api_obj = cls(address)
        await api_obj._init()
        return api_obj

    @alru_cache(cache_exceptions=False)
    async def _get_node_info_ref(self):
        from ..supervisor.node_info import NodeInfoCollectorActor

        [node_info_ref] = await self.get_supervisor_refs(
            [NodeInfoCollectorActor.default_uid()]
        )
        return node_info_ref

    async def get_supervisors(self, filter_ready: bool = True) -> List[str]:
        return await self._locator_ref.get_supervisors(filter_ready=filter_ready)

    @watch_method
    async def watch_supervisors(self, version: Optional[int] = None):
        return await self._locator_ref.watch_supervisors(version=version)

    async def get_supervisors_by_keys(self, keys: List[str]) -> List[str]:
        """
        Get supervisor address hosting the specified key

        Parameters
        ----------
        keys
            key for a supervisor address

        Returns
        -------
        out
            addresses of the supervisor
        """
        get_supervisor = self._locator_ref.get_supervisor
        return await get_supervisor.batch(*(get_supervisor.delay(k) for k in keys))

    @watch_method
    async def watch_supervisors_by_keys(
        self, keys: List[str], version: Optional[int] = None
    ):
        return await self._locator_ref.watch_supervisors_by_keys(keys, version=version)

    async def get_supervisor_refs(self, uids: List[str]) -> List[mo.ActorRef]:
        """
        Get actor references hosting the specified actor uid

        Parameters
        ----------
        uids
            uids for a supervisor address
        watch
            if True, will watch changes of supervisor changes

        Returns
        -------
        out : List[mo.ActorRef]
            references of the actors
        """
        addrs = await self.get_supervisors_by_keys(uids)
        if any(addr is None for addr in addrs):
            none_uid = next(uid for addr, uid in zip(addrs, uids) if addr is None)
            raise mo.ActorNotExist(f"Actor {none_uid} not exist as no supervisors")

        return await asyncio.gather(
            *[mo.actor_ref(uid, address=addr) for addr, uid in zip(addrs, uids)]
        )

    async def watch_supervisor_refs(self, uids: List[str]):
        async for addrs in self.watch_supervisors_by_keys(uids):
            yield await asyncio.gather(
                *[mo.actor_ref(uid, address=addr) for addr, uid in zip(addrs, uids)]
            )

    @watch_method
    async def watch_nodes(
        self,
        role: NodeRole,
        env: bool = False,
        resource: bool = False,
        detail: bool = False,
        version: Optional[int] = None,
        statuses: Set[NodeStatus] = None,
        exclude_statuses: Set[NodeStatus] = None,
    ) -> List[Dict[str, Dict]]:
        statuses = self._calc_statuses(statuses, exclude_statuses)
        node_info_ref = await self._get_node_info_ref()
        return await node_info_ref.watch_nodes(
            role,
            env=env,
            resource=resource,
            detail=detail,
            statuses=statuses,
            version=version,
        )

    async def get_nodes_info(
        self,
        nodes: List[str] = None,
        role: NodeRole = None,
        env: bool = False,
        resource: bool = False,
        detail: bool = False,
        statuses: Set[NodeStatus] = None,
        exclude_statuses: Set[NodeStatus] = None,
    ) -> Dict[str, Dict]:
        statuses = self._calc_statuses(statuses, exclude_statuses)
        node_info_ref = await self._get_node_info_ref()
        return await node_info_ref.get_nodes_info(
            nodes=nodes,
            role=role,
            env=env,
            resource=resource,
            detail=detail,
            statuses=statuses,
        )

    async def set_node_status(self, node: str, role: NodeRole, status: NodeStatus):
        """
        Set status of node

        Parameters
        ----------
        node : str
            address of node
        role: NodeRole
            role of node
        status : NodeStatus
            status of node
        """
        node_info_ref = await self._get_node_info_ref()
        await node_info_ref.update_node_info(node, role, status=status)

    async def get_all_bands(
        self,
        role: NodeRole = None,
        statuses: Set[NodeStatus] = None,
        exclude_statuses: Set[NodeStatus] = None,
    ) -> Dict[BandType, Resource]:
        statuses = self._calc_statuses(statuses, exclude_statuses)
        node_info_ref = await self._get_node_info_ref()
        return await node_info_ref.get_all_bands(role, statuses=statuses)

    @watch_method
    async def watch_all_bands(
        self,
        role: NodeRole = None,
        version: Optional[int] = None,
        statuses: Set[NodeStatus] = None,
        exclude_statuses: Set[NodeStatus] = None,
    ):
        statuses = self._calc_statuses(statuses, exclude_statuses)
        node_info_ref = await self._get_node_info_ref()
        return await node_info_ref.watch_all_bands(
            role, statuses=statuses, version=version
        )

    async def get_mars_versions(self) -> List[str]:
        node_info_ref = await self._get_node_info_ref()
        return await node_info_ref.get_mars_versions()

    async def get_bands(self) -> Dict:
        """
        Get bands that can be used for computation on current node.

        Returns
        -------
        band_to_resource : dict
            Band to resource.
        """
        return await self._uploader_ref.get_bands()

    async def mark_node_ready(self):
        """
        Mark current node ready for work loads
        """
        await self._uploader_ref.mark_node_ready()

    async def wait_node_ready(self):
        """
        Wait current node to be ready
        """
        await self._uploader_ref.wait_node_ready()

    async def wait_all_supervisors_ready(self):
        """
        Wait till all expected supervisors are ready
        """
        await self._locator_ref.wait_all_supervisors_ready()

    async def set_band_slot_infos(
        self, band_name: str, slot_infos: List[WorkerSlotInfo]
    ):
        await self._uploader_ref.set_band_slot_infos.tell(band_name, slot_infos)

    async def set_band_quota_info(self, band_name: str, quota_info: QuotaInfo):
        await self._uploader_ref.set_band_quota_info.tell(band_name, quota_info)

    async def set_node_disk_info(self, disk_info: List[DiskInfo]):
        await self._uploader_ref.set_node_disk_info(disk_info)

    @mo.extensible
    async def set_band_storage_info(self, band_name: str, storage_info: StorageInfo):
        await self._uploader_ref.set_band_storage_info(band_name, storage_info)

    async def request_worker(
        self, worker_cpu: int = None, worker_mem: int = None, timeout: int = None
    ) -> str:
        node_allocator_ref = await self._get_node_allocator_ref()
        address = await node_allocator_ref.request_worker(
            worker_cpu, worker_mem, timeout
        )
        return address

    async def release_worker(self, address: str):
        node_allocator_ref = await self._get_node_allocator_ref()
        await node_allocator_ref.release_worker(address)
        node_info_ref = await self._get_node_info_ref()
        await node_info_ref.update_node_info(
            address, NodeRole.WORKER, status=NodeStatus.STOPPED
        )

    async def reconstruct_worker(self, address: str):
        node_allocator_ref = await self._get_node_allocator_ref()
        await node_allocator_ref.reconstruct_worker(address)

    @alru_cache(cache_exceptions=False)
    async def _get_node_allocator_ref(self):
        from ..supervisor.node_allocator import NodeAllocatorActor

        [node_allocator_ref] = await self.get_supervisor_refs(
            [NodeAllocatorActor.default_uid()]
        )
        return node_allocator_ref

    async def _get_process_info_manager_ref(self, address: str = None):
        from ..procinfo import ProcessInfoManagerActor

        return await mo.actor_ref(
            ProcessInfoManagerActor.default_uid(), address=address or self._address
        )

    async def get_node_pool_configs(self, address: str = None) -> List[Dict]:
        ref = await self._get_process_info_manager_ref(address)
        return await ref.get_pool_configs()

    async def get_node_thread_stacks(
        self, address: str = None
    ) -> List[Dict[int, List[str]]]:
        ref = await self._get_process_info_manager_ref(address)
        return await ref.get_thread_stacks()


class MockClusterAPI(ClusterAPI):
    @classmethod
    async def create(cls: Type[APIType], address: str, **kw) -> APIType:
        from ..procinfo import ProcessInfoManagerActor
        from ..supervisor.locator import SupervisorPeerLocatorActor
        from ..supervisor.node_allocator import NodeAllocatorActor
        from ..supervisor.node_info import NodeInfoCollectorActor
        from ..uploader import NodeInfoUploaderActor

        create_actor_coros = [
            mo.create_actor(
                SupervisorPeerLocatorActor,
                "fixed",
                address,
                uid=SupervisorPeerLocatorActor.default_uid(),
                address=address,
            ),
            mo.create_actor(
                NodeInfoCollectorActor,
                uid=NodeInfoCollectorActor.default_uid(),
                address=address,
            ),
            mo.create_actor(
                NodeAllocatorActor,
                "fixed",
                address,
                uid=NodeAllocatorActor.default_uid(),
                address=address,
            ),
            mo.create_actor(
                NodeInfoUploaderActor,
                NodeRole.WORKER,
                interval=kw.get("upload_interval"),
                band_to_resource=kw.get("band_to_resource"),
                use_gpu=kw.get("use_gpu", False),
                uid=NodeInfoUploaderActor.default_uid(),
                address=address,
            ),
            mo.create_actor(
                ProcessInfoManagerActor,
                uid=ProcessInfoManagerActor.default_uid(),
                address=address,
            ),
        ]
        dones, _ = await asyncio.wait(
            [asyncio.ensure_future(coro) for coro in create_actor_coros]
        )

        for task in dones:
            try:
                task.result()
            except mo.ActorAlreadyExist:  # pragma: no cover
                pass

        api = await super().create(address=address)
        await api.mark_node_ready()
        return api

    @classmethod
    async def cleanup(cls, address: str):
        from ..supervisor.locator import SupervisorPeerLocatorActor
        from ..uploader import NodeInfoUploaderActor
        from ..supervisor.node_info import NodeInfoCollectorActor

        await asyncio.gather(
            mo.destroy_actor(
                mo.create_actor_ref(
                    uid=SupervisorPeerLocatorActor.default_uid(), address=address
                )
            ),
            mo.destroy_actor(
                mo.create_actor_ref(
                    uid=NodeInfoCollectorActor.default_uid(), address=address
                )
            ),
            mo.destroy_actor(
                mo.create_actor_ref(
                    uid=NodeInfoUploaderActor.default_uid(), address=address
                )
            ),
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import KeyField
from ...utils import has_unknown_shape
from ..utils import broadcast_shape, unify_chunks
from ..array_utils import as_same_device, device
from ..core import TENSOR_TYPE
from ..operands import TensorOperand, TensorOperandMixin
from ..datasource import tensor as astensor
from .broadcast_to import broadcast_to


class TensorWhere(TensorOperand, TensorOperandMixin):
    _op_type_ = OperandDef.WHERE

    _condition = KeyField("condition")
    _x = KeyField("x")
    _y = KeyField("y")

    @property
    def condition(self):
        return self._condition

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._condition = self._inputs[0]
        self._x = self._inputs[1]
        self._y = self._inputs[2]

    def __call__(self, condition, x, y, shape=None):
        shape = shape or broadcast_shape(condition.shape, x.shape, y.shape)
        return self.new_tensor([condition, x, y], shape)

    @classmethod
    def tile(cls, op):
        if has_unknown_shape(*op.inputs):
            yield
        inputs = yield from unify_chunks(
            *[(input, list(range(input.ndim))[::-1]) for input in op.inputs]
        )
        chunk_shapes = [
            t.chunk_shape if isinstance(t, TENSOR_TYPE) else t for t in inputs
        ]
        out_chunk_shape = broadcast_shape(*chunk_shapes)
        output = op.outputs[0]

        out_chunks = []
        nsplits = [[np.nan] * shape for shape in out_chunk_shape]
        get_index = lambda idx, t: tuple(
            0 if t.nsplits[i] == (1,) else ix for i, ix in enumerate(idx)
        )
        for out_index in itertools.product(*(map(range, out_chunk_shape))):
            in_chunks = [
                t.cix[get_index(out_index[-t.ndim :], t)]
                if t.ndim != 0
                else t.chunks[0]
                for t in inputs
            ]
            chunk_shape = broadcast_shape(*(c.shape for c in in_chunks))
            out_chunk = (
                op.copy()
                .reset_key()
                .new_chunk(
                    in_chunks, shape=chunk_shape, index=out_index, order=output.order
                )
            )
            out_chunks.append(out_chunk)
            for i, idx, s in zip(itertools.count(0), out_index, out_chunk.shape):
                nsplits[i][idx] = s

        new_op = op.copy()
        return new_op.new_tensors(
            inputs, output.shape, order=output.order, chunks=out_chunks, nsplits=nsplits
        )

    @classmethod
    def execute(cls, ctx, op):
        (cond, x, y), device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True
        )

        with device(device_id):
            ctx[op.outputs[0].key] = xp.where(cond, x, y)


def where(condition, x=None, y=None):
    """
    Return elements, either from `x` or `y`, depending on `condition`.

    If only `condition` is given, return ``condition.nonzero()``.

    Parameters
    ----------
    condition : array_like, bool
        When True, yield `x`, otherwise yield `y`.
    x, y : array_like, optional
        Values from which to choose. `x`, `y` and `condition` need to be
        broadcastable to some shape.

    Returns
    -------
    out : Tensor or tuple of Tensors
        If both `x` and `y` are specified, the output tensor contains
        elements of `x` where `condition` is True, and elements from
        `y` elsewhere.

        If only `condition` is given, return the tuple
        ``condition.nonzero()``, the indices where `condition` is True.

    See Also
    --------
    nonzero, choose

    Notes
    -----
    If `x` and `y` are given and input arrays are 1-D, `where` is
    equivalent to::

        [xv if c else yv for (c,xv,yv) in zip(condition,x,y)]

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.where([[True, False], [True, True]],
    ...          [[1, 2], [3, 4]],
    ...          [[9, 8], [7, 6]]).execute()
    array([[1, 8],
           [3, 4]])

    >>> mt.where([[0, 1], [1, 0]]).execute()
    (array([0, 1]), array([1, 0]))

    >>> x = mt.arange(9.).reshape(3, 3)
    >>> mt.where( x > 5 ).execute()
    (array([2, 2, 2]), array([0, 1, 2]))
    >>> mt.where(x < 5, x, -1).execute()               # Note: broadcasting.
    array([[ 0.,  1.,  2.],
           [ 3.,  4., -1.],
           [-1., -1., -1.]])

    Find the indices of elements of `x` that are in `goodvalues`.

    >>> goodvalues = [3, 4, 7]
    >>> ix = mt.isin(x, goodvalues)
    >>> ix.execute()
    array([[False, False, False],
           [ True,  True, False],
           [False,  True, False]])
    >>> mt.where(ix).execute()
    (array([1, 1, 2]), array([0, 1, 1]))
    """
    if (x is None) != (y is None):
        raise ValueError("either both or neither of x and y should be given")

    if x is None and y is None:
        return astensor(condition).nonzero()

    x, y = astensor(x), astensor(y)
    dtype = np.result_type(x.dtype, y.dtype)
    shape = broadcast_shape(x.shape, y.shape)

    if np.isscalar(condition):
        return broadcast_to(x if condition else y, shape).astype(dtype)
    else:
        condition = astensor(condition)
        op = TensorWhere(dtype=dtype)
        return op(condition, x, y, shape=shape)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio.tasks
import contextvars
import json
import logging
import os
import time
from contextlib import contextmanager
from dataclasses import dataclass
from typing import List, Optional  # noqa: F401

from ..utils import dataslots

logger: logging.Logger = logging.getLogger(__name__)


@dataslots
@dataclass
class MessageTraceItem:
    uid: str
    address: str
    method: str


@dataslots
@dataclass
class DebugOptions:
    actor_call_timeout: int = 10
    process_message_timeout: int = 30
    actor_lock_timeout: int = 30
    ray_object_retrieval_timeout: int = 10
    log_unhandled_errors: bool = True
    log_cycle_send: bool = True


_debug_opts: Optional[DebugOptions] = None


def get_debug_options() -> Optional[DebugOptions]:
    return _debug_opts


def set_debug_options(options: Optional[DebugOptions]):
    global _debug_opts
    _debug_opts = options

    # deliver debug config to native codes for optimization
    from .core import set_debug_options as core_set_debug_options

    core_set_debug_options(options)


def reload_debug_opts_from_env():
    config_str = os.environ.get("DEBUG_OSCAR", "0")
    if config_str == "0":
        set_debug_options(None)
        return
    config_str = os.environ["DEBUG_OSCAR"]
    config_json = {} if config_str == "1" else json.loads(config_str)
    set_debug_options(DebugOptions(**config_json))


async def _log_timeout(timeout, msg, *args, **kwargs):
    start_time, rnd = time.time(), 1
    while True:
        await asyncio.sleep(timeout * rnd)
        rnd += 1
        logger.warning(
            msg + " (timeout for %.4f seconds).",
            *args,
            time.time() - start_time,
            **kwargs,
        )


@contextmanager
def debug_async_timeout(option_name: str, msg, *args, **kwargs):
    if _debug_opts is None:
        yield
    else:
        timeout_val = getattr(_debug_opts, option_name, -1)
        timeout_task = None
        if timeout_val and timeout_val > 0:
            timeout_task = asyncio.create_task(
                _log_timeout(timeout_val, msg, *args, **kwargs)
            )

        try:
            yield
        finally:
            if timeout_task is not None:
                timeout_task.cancel()


_message_trace_var = contextvars.ContextVar("_message_trace_var")


@contextmanager
def record_message_trace(message):
    if _debug_opts is None or not _debug_opts.log_cycle_send:
        yield
    else:
        msg_trace = list(message.message_trace or [])
        msg_trace.append(
            MessageTraceItem(
                uid=message.actor_ref.uid,
                address=message.actor_ref.address,
                method=message.content[0],
            )
        )
        _message_trace_var.set(msg_trace)
        try:
            yield
        finally:
            _message_trace_var.set(None)


def detect_cycle_send(message, wait_response: bool = True):
    if _debug_opts is None or not _debug_opts.log_cycle_send or not wait_response:
        return

    cur_trace = _message_trace_var.get(None) or []  # type: List[MessageTraceItem]
    message.message_trace = cur_trace

    ref_key = (message.actor_ref.uid, message.actor_ref.address)
    traced_ref_keys = set((item.uid, item.address) for item in cur_trace)
    if ref_key in traced_ref_keys:
        looped_trace = cur_trace + [
            MessageTraceItem(
                uid=message.actor_ref.uid,
                address=message.actor_ref.address,
                method=message.content[0],
            )
        ]

        formatted_trace = "\n    ".join(
            f"Calling {t.method!r} in actor {t.uid} at {t.address}"
            for t in looped_trace
        )
        logger.warning(
            "Call cycle detected when sending to actor %s at %s, the trace is\n"
            "    %s",
            message.actor_ref.uid,
            message.actor_ref.address,
            formatted_trace,
        )


@contextmanager
def no_message_trace():
    if _debug_opts is None or not _debug_opts.log_cycle_send:
        yield
    else:
        trace = pop_message_trace()
        yield
        set_message_trace(trace)


def pop_message_trace():
    trace = _message_trace_var.get(None)
    _message_trace_var.set(None)
    return trace


def set_message_trace(message_trace):
    _message_trace_var.set(message_trace)


reload_debug_opts_from_env()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import OutputType
from ...serialization.serializables import AnyField, DictField, StringField, Int64Field
from ..utils import parse_index, build_empty_df, build_empty_series
from ..operands import DataFrameOperandMixin, DataFrameOperand


class GroupByFillOperand(DataFrameOperand, DataFrameOperandMixin):
    _op_module_ = "dataframe.groupby"

    value = AnyField("value", default=None)
    method = StringField("method", default=None)
    axis = AnyField("axis", default=0)
    limit = Int64Field("limit", default=None)
    downcast = DictField("downcast", default=None)

    def _calc_out_dtypes(self, in_groupby):
        mock_groupby = in_groupby.op.build_mock_groupby()
        func_name = getattr(self, "_func_name")

        if func_name == "fillna":
            result_df = mock_groupby.fillna(
                value=self.value,
                method=self.method,
                axis=self.axis,
                limit=self.limit,
                downcast=self.downcast,
            )
        else:
            result_df = getattr(mock_groupby, func_name)(limit=self.limit)

        if isinstance(result_df, pd.DataFrame):
            self.output_types = [OutputType.dataframe]
            return result_df.dtypes
        else:
            self.output_types = [OutputType.series]
            return result_df.name, result_df.dtype

    def __call__(self, groupby):
        in_df = groupby
        while in_df.op.output_types[0] not in (OutputType.dataframe, OutputType.series):
            in_df = in_df.inputs[0]
        out_dtypes = self._calc_out_dtypes(groupby)

        kw = in_df.params.copy()
        kw["index_value"] = parse_index(pd.RangeIndex(-1), groupby.key)
        if self.output_types[0] == OutputType.dataframe:
            kw.update(
                dict(
                    columns_value=parse_index(out_dtypes.index, store_data=True),
                    dtypes=out_dtypes,
                    shape=(groupby.shape[0], len(out_dtypes)),
                )
            )
        else:
            name, dtype = out_dtypes
            kw.update(dtype=dtype, name=name, shape=(groupby.shape[0],))
        return self.new_tileable([groupby], **kw)

    @classmethod
    def tile(cls, op):
        in_groupby = op.inputs[0]
        out_df = op.outputs[0]

        chunks = []
        for c in in_groupby.chunks:
            new_op = op.copy().reset_key()

            new_index = parse_index(pd.RangeIndex(-1), c.key)
            if op.output_types[0] == OutputType.dataframe:
                chunks.append(
                    new_op.new_chunk(
                        [c],
                        index=c.index,
                        shape=(np.nan, len(out_df.dtypes)),
                        dtypes=out_df.dtypes,
                        columns_value=out_df.columns_value,
                        index_value=new_index,
                    )
                )
            else:
                chunks.append(
                    new_op.new_chunk(
                        [c],
                        index=(c.index[0],),
                        shape=(np.nan,),
                        dtype=out_df.dtype,
                        index_value=new_index,
                        name=out_df.name,
                    )
                )
        new_op = op.copy().reset_key()
        kw = out_df.params.copy()
        kw["chunks"] = chunks
        if op.output_types[0] == OutputType.dataframe:
            kw["nsplits"] = ((np.nan,) * len(chunks), (len(out_df.dtypes),))
        else:
            kw["nsplits"] = ((np.nan,) * len(chunks),)
        return new_op.new_tileables([in_groupby], **kw)

    @classmethod
    def execute(cls, ctx, op: "GroupByFillOperand"):
        in_data = ctx[op.inputs[0].key]
        out_chunk = op.outputs[0]

        if not in_data or in_data.empty:
            ctx[out_chunk.key] = (
                build_empty_df(out_chunk.dtypes)
                if op.output_types[0] == OutputType.dataframe
                else build_empty_series(out_chunk.dtype, name=out_chunk.name)
            )
            return

        func_name = getattr(op, "_func_name")
        if func_name == "fillna":
            ctx[out_chunk.key] = in_data.fillna(
                value=op.value,
                method=op.method,
                axis=op.axis,
                limit=op.limit,
                downcast=op.downcast,
            )
        else:
            result = getattr(in_data, func_name)(limit=op.limit)
            if result.ndim == 2:
                ctx[out_chunk.key] = result.astype(out_chunk.dtypes, copy=False)
            else:
                ctx[out_chunk.key] = result.astype(out_chunk.dtype, copy=False)


class GroupByFFill(GroupByFillOperand):
    _op_type_ = opcodes.FILL_NA
    _func_name = "ffill"


class GroupByBFill(GroupByFillOperand):
    _op_type = opcodes.FILL_NA
    _func_name = "bfill"


class GroupByFillNa(GroupByFillOperand):
    _op_type = opcodes.FILL_NA
    _func_name = "fillna"


def ffill(groupby, limit=None):
    """
    Forward fill the values.

    limit:  int, default None
            Limit number of values to fill

    return: Series or DataFrame
    """
    op = GroupByFFill(limit=limit)
    return op(groupby)


def bfill(groupby, limit=None):
    """
    Backward fill the values.

    limit:  int, default None
            Limit number of values to fill

    return: Series or DataFrame
    """
    op = GroupByBFill(limit=limit)
    return op(groupby)


def fillna(groupby, value=None, method=None, axis=None, limit=None, downcast=None):
    """
    Fill NA/NaN values using the specified method

    value:  scalar, dict, Series, or DataFrame
            Value to use to fill holes (e.g. 0), alternately a dict/Series/DataFrame
            of values specifying which value to use for each index (for a Series) or
            column (for a DataFrame). Values not in the dict/Series/DataFrame
            will not be filled. This value cannot be a list.
    method: {'backfill','bfill','ffill',None}, default None
    axis:   {0 or 'index', 1 or 'column'}
    limit:  int, default None
            If method is specified, this is the maximum number of consecutive
            NaN values to forward/backward fill
    downcast:   dict, default None
                A dict of item->dtype of what to downcast if possible,
                or the string ‘infer’ which will try to downcast to an appropriate equal type

    return: DataFrame or None
    """
    op = GroupByFillNa(
        value=value, method=method, axis=axis, limit=limit, downcast=downcast
    )
    return op(groupby)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import functools
import logging
import os
import time
from typing import Dict, List, Optional

from .... import oscar as mo
from ....utils import to_binary
from ...cluster import ClusterAPI
from ...core import NodeRole, create_service_session, destroy_service_session
from ..core import SessionInfo

logger = logging.getLogger(__name__)


class SessionManagerActor(mo.Actor):
    def __init__(self, service_config: Optional[Dict] = None):
        self._session_refs: Dict[str, mo.ActorRef] = dict()
        self._cluster_api: Optional[ClusterAPI] = None
        self._service_config = service_config or dict()
        self._stored_last_idle_time = None

    async def __post_create__(self):
        self._cluster_api = await ClusterAPI.create(self.address)
        self._stored_last_idle_time = time.time()

    async def __pre_destroy__(self):
        await asyncio.gather(
            *[mo.destroy_actor(ref) for ref in self._session_refs.values()]
        )

    async def create_session(self, session_id: str, create_services: bool = True):
        if session_id in self._session_refs:
            raise mo.Return(self._session_refs[session_id])

        [address] = await self._cluster_api.get_supervisors_by_keys([session_id])
        try:
            session_actor_ref = await mo.create_actor(
                SessionActor,
                session_id,
                self._service_config,
                address=address,
                uid=SessionActor.gen_uid(session_id),
                allocate_strategy=mo.allocate_strategy.RandomSubPool(),
            )
        except IndexError:
            # when there is only one supervisor process, strategy RandomSubPool
            # fails with IndexError. So we need to retry using strategy Random.
            session_actor_ref = await mo.create_actor(
                SessionActor,
                session_id,
                self._service_config,
                address=address,
                uid=SessionActor.gen_uid(session_id),
                allocate_strategy=mo.allocate_strategy.Random(),
            )
        self._session_refs[session_id] = session_actor_ref

        # sync ref to other managers
        for supervisor_address in await self._cluster_api.get_supervisors():
            if supervisor_address == self.address:
                continue
            session_manager_ref = await mo.actor_ref(
                supervisor_address, SessionManagerActor.default_uid()
            )
            await session_manager_ref.add_session_ref(session_id, session_actor_ref)

        # let session actor create session-related services
        if create_services:
            yield session_actor_ref.create_services()

        raise mo.Return(session_actor_ref)

    def get_sessions(self) -> List[SessionInfo]:
        return [
            SessionInfo(session_id=session_id)
            for session_id in self._session_refs.keys()
        ]

    def get_session_ref(self, session_id: str):
        return self._session_refs[session_id]

    def add_session_ref(self, session_id: str, session_actor_ref: mo.ActorRef):
        self._session_refs[session_id] = session_actor_ref

    def remove_session_ref(self, session_id: str):
        del self._session_refs[session_id]

    def has_session(self, session_id: str):
        return session_id in self._session_refs

    async def delete_session(self, session_id):
        session_actor_ref = self._session_refs.pop(session_id)
        await session_actor_ref.remove()
        await mo.destroy_actor(session_actor_ref)

        # sync removing to other managers
        for supervisor_address in await self._cluster_api.get_supervisors():
            if supervisor_address == self.address:
                continue
            session_manager_ref = await mo.actor_ref(
                supervisor_address, SessionManagerActor.default_uid()
            )
            await session_manager_ref.remove_session_ref(session_id)

    async def delete_all_sessions(self):
        for session_id in list(self._session_refs):
            await self.delete_session(session_id)

    async def get_last_idle_time(self, session_id=None):
        if session_id is not None:
            session = self._session_refs[session_id]
            raise mo.Return(await session.get_last_idle_time())
        else:
            all_last_idle_time = yield asyncio.gather(
                *[
                    session.get_last_idle_time()
                    for session in self._session_refs.values()
                ]
            )
            if any(last_idle_time is None for last_idle_time in all_last_idle_time):
                raise mo.Return(None)
            else:
                self._stored_last_idle_time = max(
                    [self._stored_last_idle_time] + all_last_idle_time
                )
                raise mo.Return(self._stored_last_idle_time)


class SessionActor(mo.Actor):
    def __init__(self, session_id: str, service_config: Dict):
        self._session_id = session_id

        self._meta_api = None
        self._lifecycle_api = None
        self._task_api = None
        self._scheduling_api = None

        self._service_config = service_config

        self._custom_log_meta_ref = None

    @classmethod
    def gen_uid(cls, session_id):
        return f"{session_id}_session_actor"

    async def __post_create__(self):
        from .custom_log import CustomLogMetaActor

        self._custom_log_meta_ref = await mo.create_actor(
            CustomLogMetaActor,
            self._session_id,
            address=self.address,
            uid=CustomLogMetaActor.gen_uid(self._session_id),
        )
        logger.debug(
            "Session %s actor created on pid: %s",
            self._session_id,
            os.getpid(),
        )

    async def remove(self):
        await destroy_service_session(
            NodeRole.SUPERVISOR, self._service_config, self._session_id, self.address
        )

    async def __pre_destroy__(self):
        await mo.destroy_actor(self._custom_log_meta_ref)

    async def create_services(self):
        from ...task import TaskAPI

        await create_service_session(
            NodeRole.SUPERVISOR, self._service_config, self._session_id, self.address
        )
        if "task" in self._service_config["services"]:
            self._task_api = await TaskAPI.create(
                session_id=self._session_id, address=self.address
            )

    async def get_last_idle_time(self):
        if self._task_api is None:
            return None
        return await self._task_api.get_last_idle_time()

    async def create_remote_object(self, name: str, object_cls, *args, **kwargs):
        return await mo.create_actor(
            RemoteObjectActor,
            object_cls,
            args,
            kwargs,
            address=self.address,
            uid=to_binary(name),
        )

    async def get_remote_object(self, name: str):
        return await mo.actor_ref(mo.ActorRef(self.address, to_binary(name)))

    async def destroy_remote_object(self, name: str):
        return await mo.destroy_actor(mo.ActorRef(self.address, to_binary(name)))


class RemoteObjectActor(mo.Actor):
    def __init__(self, object_cls, args, kwargs):
        self._object = object_cls(*args, **kwargs)

    def __getattr__(self, attr):
        func = getattr(self._object, attr)
        if not callable(func):  # pragma: no cover
            return object.__getattribute__(self._object, attr)

        @functools.wraps(func)
        async def wrap(*args, **kwargs):
            # return coroutine to not block current actor
            if asyncio.iscoroutinefunction(func):
                return func(*args, **kwargs)
            else:
                # for sync call, running in thread
                return asyncio.to_thread(func, *args, **kwargs)

        return wrap



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
from collections import defaultdict
from typing import Dict, Tuple

from .... import oscar as mo


class CustomLogMetaActor(mo.Actor):
    # {tileable_op_key -> {chunk_op_key -> (worker_addr, path)}}
    _custom_log_path_store: Dict[str, Dict[str, Tuple[str, str]]]

    def __init__(self, session_id: str):
        self._session_id = session_id
        self._custom_log_path_store = dict()

    @classmethod
    def gen_uid(cls, session_id: str):
        return f"custom_log_{session_id}"

    async def __post_create__(self):
        from ..worker.custom_log import CustomLogActor

        worker_address_to_paths = defaultdict(set)
        for address, path in self._custom_log_path_store.values():
            log_dir = os.path.dirname(path)
            worker_address_to_paths[address].add(log_dir)
        for address, paths in worker_address_to_paths.items():
            ref = await mo.actor_ref(address, CustomLogActor.default_uid())
            await ref.clear_custom_log_dirs(list(paths))

    def register_custom_log_path(
        self,
        tileable_op_key: str,
        chunk_op_key: str,
        worker_address: str,
        log_path: str,
    ):
        if tileable_op_key not in self._custom_log_path_store:
            self._custom_log_path_store[tileable_op_key] = dict()
        self._custom_log_path_store[tileable_op_key][chunk_op_key] = (
            worker_address,
            log_path,
        )

    def get_tileable_op_log_paths(
        self, tileable_op_key: str
    ) -> Dict[str, Tuple[str, str]]:
        return self._custom_log_path_store.get(tileable_op_key)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from collections import namedtuple
from typing import List, Union, Tuple

import numpy as np
import pandas as pd
from pandas.core.dtypes.cast import find_common_type

from ...core import Tileable, Chunk, OutputType, recursive_tile
from ...core.operand import OperandStage
from ...tensor.core import TENSOR_TYPE
from ...tensor.indexing.index_lib import (
    IndexHandlerContext,
    IndexHandler,
    IndexInfo,
    IndexType,
    ChunkIndexInfo as ChunkIndexInfoBase,
    SliceIndexHandler as SliceIndexHandlerBase,
    NDArrayBoolIndexHandler as NDArrayBoolIndexHandlerBase,
    TensorBoolIndexHandler as TensorBoolIndexHandlerBase,
    IntegralIndexHandler,
    IndexesHandler,
)
from ...tensor.utils import (
    split_indexes_into_chunks,
    calc_pos,
    filter_inputs,
    slice_split,
    calc_sliced_size,
    to_numpy,
    normalize_chunk_sizes,
)
from ...utils import classproperty, has_unknown_shape, is_full_slice
from ..core import SERIES_CHUNK_TYPE, SERIES_TYPE, IndexValue
from ..utils import parse_index
from .utils import convert_labels_into_positions


ChunkIndexAxisInfo = namedtuple(
    "chunk_index_axis_info",
    ["output_axis_index", "processed_index", "output_shape", "index_value", "dtypes"],
)


class ChunkIndexInfo(ChunkIndexInfoBase):
    def __init__(self):
        super().__init__()
        self.index_values = []
        self.dtypes = None

    def set(self, info: ChunkIndexAxisInfo):
        super().set(info)
        if getattr(info, "index_value", None) is not None:
            self.index_values.append(info.index_value)
        if getattr(info, "dtypes", None) is not None:
            self.dtypes = info.dtypes


class FancyIndexInfo(IndexInfo):
    def __init__(
        self,
        index_type: IndexType,
        input_axis: int,
        output_axis: int,
        raw_index,
        handler,
    ):
        super().__init__(index_type, input_axis, output_axis, raw_index, handler)

        # extra info for DataFrame fancy index
        # split info
        #   - chunk_index_to_fancy_index_arrays
        #   - chunk_index_to_raw_positions
        #   - is_fancy_index_asc_sorted
        self.split_info = None


class LabelFancyIndexInfo(IndexInfo):
    def __init__(
        self,
        index_type: IndexType,
        input_axis: int,
        output_axis: int,
        raw_index,
        handler,
    ):
        super().__init__(index_type, input_axis, output_axis, raw_index, handler)

        # store chunk_index -> labels
        self.chunk_index_to_labels = None
        self.is_label_asc_sorted = None


class DataFrameIndexHandlerContext(IndexHandlerContext):
    def set_tileable(self, tileable: Tileable):
        for chunk in tileable.chunks:
            self.chunk_index_to_info[chunk.index] = ChunkIndexInfo()

    def concat_chunks(self, chunks: List[Chunk], axis: Union[Tuple[int], int]) -> Chunk:
        dataframe_op_type = type(chunks[0].op)
        # create tileable from chunks
        concat_tileable = dataframe_op_type.create_tileable_from_chunks(
            chunks, inputs=chunks
        )
        # concat chunks
        chunk = dataframe_op_type.concat_tileable_chunks(concat_tileable).chunks[0]
        if chunk.ndim > 1 and (
            (isinstance(axis, tuple) and len(axis) == 1) or isinstance(axis, int)
        ):
            # adjust index and axis
            axis = axis[0] if isinstance(axis, tuple) else axis
            chunk.op._axis = axis
            chunk_index = list(chunk.index)
            chunk_index[1 - axis] = chunks[0].index[1 - axis]
            chunk._index = tuple(chunk_index)
        return chunk

    def create_chunk(
        self, chunk_index: Tuple[int], chunk_index_info: ChunkIndexInfo
    ) -> Chunk:
        chunk_op = self.op.copy().reset_key()
        chunk_op._indexes = indexes = chunk_index_info.indexes
        chunk_op.stage = OperandStage.map

        chunk_input = self.tileable.cix[chunk_index]
        chunk_inputs = filter_inputs([chunk_input] + indexes)

        kw = {}
        kw["shape"] = shape = tuple(chunk_index_info.output_chunk_shape)
        kw["index"] = tuple(chunk_index_info.output_chunk_index)
        index_values = chunk_index_info.index_values
        if len(shape) == 0:
            # scalar
            chunk_op.output_types = [OutputType.scalar]
            kw["dtype"] = self.op.outputs[0].dtype
        elif len(shape) == 1:
            # Series or Index
            chunk_op.output_types = (
                [OutputType.index]
                if chunk_op._op_module_ == "index"
                else [OutputType.series]
            )
            kw["index_value"] = index_values[0]
            kw["dtype"] = self.op.outputs[0].dtype
            kw["name"] = getattr(self.op.outputs[0], "name", None)
        else:
            # dataframe
            chunk_op.output_types = [OutputType.dataframe]
            kw["index_value"] = index_values[0]
            kw["columns_value"] = index_values[1]
            kw["dtypes"] = chunk_index_info.dtypes

        return chunk_op.new_chunk(chunk_inputs, kws=[kw])


class SliceIndexHandler(SliceIndexHandlerBase):
    @classmethod
    def set_chunk_index_info(
        cls,
        context: IndexHandlerContext,
        index_info: IndexInfo,
        chunk_index: Tuple[int],
        chunk_index_info: ChunkIndexInfo,
        output_axis_index: int,
        index,
        output_shape: int,
    ):
        tileable = context.tileable
        chunk_input = tileable.cix[chunk_index]
        slc = index

        kw = {
            "output_axis_index": output_axis_index,
            "processed_index": slc,
            "output_shape": output_shape,
            "dtypes": None,
        }
        if index_info.input_axis == 0:
            if is_full_slice(slc):
                kw["index_value"] = chunk_input.index_value
            else:
                index = chunk_input.index_value.to_pandas()
                kw["index_value"] = parse_index(
                    index[slc], chunk_input, slc, store_data=False
                )
        else:
            assert index_info.input_axis == 1
            index = chunk_input.columns_value.to_pandas()
            # do not store index value if output axis is 0
            store_data = True if index_info.output_axis == 1 else False
            kw["index_value"] = parse_index(index[slc], store_data=store_data)
            kw["dtypes"] = chunk_input.dtypes[slc]

        chunk_index_info.set(ChunkIndexAxisInfo(**kw))


class LabelSliceIndexHandler(IndexHandler):
    def accept(cls, raw_index):
        return isinstance(raw_index, slice)

    def parse(self, raw_index, context: IndexHandlerContext) -> IndexInfo:
        info = IndexInfo(
            IndexType.label_slice,
            context.input_axis,
            context.output_axis,
            raw_index,
            self,
        )
        context.input_axis += 1
        context.output_axis += 1
        context.append(info)
        return info

    @staticmethod
    def _slice_all(slc):
        return (
            slc.start is None
            and slc.stop is None
            and (slc.step is None or slc.step == 1)
        )

    def preprocess(self, index_info: IndexInfo, context: IndexHandlerContext) -> None:
        tileable = context.tileable
        input_axis = index_info.input_axis
        if isinstance(tileable, SERIES_TYPE):
            index_value = tileable.index_value
        else:
            index_value = [tileable.index_value, tileable.columns_value][input_axis]

        # check if chunks have unknown shape
        if (
            not self._slice_all(index_info.raw_index)
            and index_value.has_value()
            and any(np.isnan(ns) for ns in tileable.nsplits[input_axis])
        ):  # pragma: no cover
            yield

    def set_chunk_index_info(
        cls,
        context: IndexHandlerContext,
        index_info: IndexInfo,
        chunk_index: Tuple[int],
        chunk_index_info: ChunkIndexInfo,
        output_axis_index: int,
        index,
        output_shape: int,
    ):
        tileable = context.tileable
        chunk_input = tileable.cix[chunk_index]
        slc = index

        kw = {
            "output_axis_index": output_axis_index,
            "processed_index": slc,
            "output_shape": output_shape,
            "dtypes": None,
        }
        if index_info.input_axis == 0:
            if is_full_slice(index):
                kw["index_value"] = chunk_input.index_value
            else:
                index = chunk_input.index_value.to_pandas()
                start, stop = index.slice_locs(
                    slc.start, slc.stop, slc.step, kind="loc"
                )
                pos_slc = slice(start, stop, slc.step)
                kw["index_value"] = parse_index(
                    index[pos_slc], chunk_input, slc, store_data=False
                )
        else:
            assert index_info.input_axis == 1
            dtypes = chunk_input.dtypes
            # do not store index value if output axis is 0
            store_data = True if index_info.output_axis == 1 else False
            columns = dtypes.loc[slc].index
            kw["index_value"] = parse_index(columns, store_data=store_data)
            kw["dtypes"] = chunk_input.dtypes[slc]

        chunk_index_info.set(ChunkIndexAxisInfo(**kw))

    def _process_slice_all_index(
        self,
        tileable: Tileable,
        index_info: IndexInfo,
        input_axis: int,
        context: IndexHandlerContext,
    ) -> None:
        index_to_info = context.chunk_index_to_info.copy()
        for chunk_index, chunk_index_info in index_to_info.items():
            i = chunk_index[input_axis]
            size = tileable.nsplits[input_axis][i]
            self.set_chunk_index_info(
                context,
                index_info,
                chunk_index,
                chunk_index_info,
                i,
                slice(None),
                size,
            )

    def _process_has_value_index(
        self,
        tileable: Tileable,
        index_info: IndexInfo,
        index_value,
        input_axis: int,
        context: IndexHandlerContext,
    ) -> None:
        pd_index = index_value.to_pandas()
        # turn label-based slice into position-based slice
        start, end = pd_index.slice_locs(
            index_info.raw_index.start,
            index_info.raw_index.stop,
            index_info.raw_index.step,
            kind="loc",
        )
        slc = slice(start, end, index_info.raw_index.step)

        cum_nsplit = [0] + np.cumsum(tileable.nsplits[index_info.input_axis]).tolist()
        # split position-based slice into chunk slices
        effected_i_to_slc = slice_split(slc, tileable.nsplits[index_info.input_axis])
        is_reversed = (slc.step or 0) < 0
        output_axis_index_range = (
            range(len(effected_i_to_slc))
            if not is_reversed
            else range(len(effected_i_to_slc) - 1, -1, -1)
        )
        other_index_to_iter = dict()

        index_to_info = context.chunk_index_to_info.copy()
        for chunk_index, chunk_index_info in index_to_info.items():
            i = chunk_index[input_axis]
            other_index = chunk_index[:input_axis] + chunk_index[input_axis + 1 :]
            size = tileable.nsplits[input_axis][i]
            if i not in effected_i_to_slc:
                # delete it, the input chunk could be ignored
                del context.chunk_index_to_info[chunk_index]
            else:
                chunk_slc = effected_i_to_slc[i]
                output_shape = calc_sliced_size(size, chunk_slc)
                if other_index not in other_index_to_iter:
                    other_index_to_iter[other_index] = iter(output_axis_index_range)
                output_axis_index = next(other_index_to_iter[other_index])

                # turn position-based slice back into label-based slice
                start = chunk_slc.start
                if start is not None:
                    abs_start = cum_nsplit[i] + start
                    label_start = pd_index[abs_start]
                else:
                    label_start = None
                stop = chunk_slc.stop
                if stop is not None:
                    abs_stop = cum_nsplit[i] + stop - 1  # label slice include the stop
                    label_stop = (
                        pd_index[abs_stop] if abs_stop < len(pd_index) else None
                    )
                else:
                    label_stop = None

                label_slc = slice(label_start, label_stop, chunk_slc.step)
                self.set_chunk_index_info(
                    context,
                    index_info,
                    chunk_index,
                    chunk_index_info,
                    output_axis_index,
                    label_slc,
                    output_shape,
                )

    def process(self, index_info: IndexInfo, context: IndexHandlerContext) -> None:
        tileable = context.tileable
        input_axis = index_info.input_axis
        if isinstance(tileable, SERIES_TYPE):
            index_value = tileable.index_value
        else:
            index_value = [tileable.index_value, tileable.columns_value][input_axis]

        if self._slice_all(index_info.raw_index):
            self._process_slice_all_index(tileable, index_info, input_axis, context)
        elif index_value.has_value():
            self._process_has_value_index(
                tileable, index_info, index_value, input_axis, context
            )
        else:
            other_index_to_iter = dict()
            # slice on all chunks on the specified axis
            for chunk_index, chunk_index_info in context.chunk_index_to_info.items():
                other_index = chunk_index[:1] if input_axis == 1 else chunk_index[1:]
                if other_index not in other_index_to_iter:
                    other_index_to_iter[other_index] = itertools.count()
                output_axis_index = next(other_index_to_iter[other_index])
                self.set_chunk_index_info(
                    context,
                    index_info,
                    chunk_index,
                    chunk_index_info,
                    output_axis_index,
                    index_info.raw_index,
                    np.nan,
                )


class LabelIndexHandler(IndexHandler):
    def accept(cls, raw_index):
        # accept type other than slice, ndarray and tensor
        return not isinstance(raw_index, (slice, np.ndarray, TENSOR_TYPE))

    def parse(self, raw_index, context: IndexHandlerContext) -> IndexInfo:
        tileable = context.tileable
        input_axis = context.input_axis
        if tileable.ndim == 2:
            index_value = [tileable.index_value, tileable.columns_value][input_axis]
        else:
            index_value = tileable.index_value

        if index_value.has_value():
            pd_index = index_value.to_pandas()
            loc = pd_index.get_loc(raw_index)
            if isinstance(loc, slice):
                # if is slice, means index not unique, but monotonic
                # just call LabelSliceIndexHandler
                new_raw_index = slice(raw_index, raw_index)
                return LabelSliceIndexHandler.get_instance().parse(
                    new_raw_index, context
                )
            elif isinstance(loc, np.ndarray):
                # bool indexing, non unique, and not monotonic
                return NDArrayBoolIndexHandler.get_instance().parse(loc, context)
        else:
            return LabelNDArrayFancyIndexHandler.get_instance().parse(
                raw_index, context
            )

        info = IndexInfo(
            IndexType.label, context.input_axis, context.output_axis, raw_index, self
        )
        context.input_axis += 1
        context.append(info)
        return info

    def preprocess(self, index_info: IndexInfo, context: IndexHandlerContext) -> None:
        # if index has value on input axis,
        # label will be converted to position,
        # thus chunks cannot have unknown shape on this axis
        tileable = context.tileable
        input_axis = index_info.input_axis
        if tileable.ndim == 1:
            index_value = tileable.index_value
        else:
            index_value = [tileable.index_value, tileable.columns_value][input_axis]
        if index_value.has_value():
            if any(
                np.isnan(ns) for ns in tileable.nsplits[input_axis]
            ):  # pragma: no cover
                yield

    def process(self, index_info: IndexInfo, context: IndexHandlerContext) -> None:
        tileable = context.tileable
        input_axis = index_info.input_axis
        if tileable.ndim == 1:
            index_value = tileable.index_value
        else:
            index_value = [tileable.index_value, tileable.columns_value][input_axis]

        if index_value.has_value():
            pd_index = index_value.to_pandas()
            loc = pd_index.get_loc(index_info.raw_index)

            # other situations have been delegated to different handlers
            assert isinstance(loc, int)

            effected_i_to_slc = slice_split(
                loc, tileable.nsplits[index_info.input_axis]
            )

            index_to_info = context.chunk_index_to_info.copy()
            for chunk_index, chunk_index_info in index_to_info.items():
                i = chunk_index[input_axis]
                if i not in effected_i_to_slc:
                    # delete it, the input chunk could be ignored
                    del context.chunk_index_to_info[chunk_index]
                else:
                    chunk_index_info.set(
                        ChunkIndexAxisInfo(
                            output_axis_index=None,
                            processed_index=index_info.raw_index,
                            output_shape=None,
                            index_value=None,
                            dtypes=None,
                        )
                    )


class DataFrameIndexHandler:
    @classmethod
    def _calc_dtypes(cls, dtypes, index, context: IndexHandlerContext):
        if getattr(context.op, "can_index_miss", False):
            # reindex
            return dtypes.reindex(index).fillna(np.dtype(np.float64))
        else:
            # loc, iloc
            return getattr(dtypes, cls.kind)[index]

    @classmethod
    def set_chunk_index_info(
        cls,
        context: IndexHandlerContext,
        index_info: IndexInfo,
        chunk_index: Tuple[int],
        chunk_index_info: ChunkIndexInfo,
        output_axis_index: int,
        index,
        output_shape: int,
    ):
        tileable = context.tileable
        chunk_input = tileable.cix[chunk_index]

        dtypes = None
        if index_info.input_axis == 0:
            index_value = parse_index(
                chunk_input.index_value.to_pandas()[:0],
                chunk_input,
                index,
                store_data=False,
            )
        else:
            dtypes = cls._calc_dtypes(chunk_input.dtypes, index, context)
            columns = dtypes.index
            index_value = parse_index(columns, store_data=True)

        info = ChunkIndexAxisInfo(
            output_axis_index=output_axis_index,
            processed_index=index,
            output_shape=output_shape,
            index_value=index_value,
            dtypes=dtypes,
        )
        chunk_index_info.set(info)


class NDArrayBoolIndexHandler(NDArrayBoolIndexHandlerBase):
    @classmethod
    def set_chunk_index_info(
        cls,
        context: IndexHandlerContext,
        index_info: IndexInfo,
        chunk_index: Tuple[int],
        chunk_index_info: ChunkIndexInfo,
        output_axis_index: int,
        index,
        output_shape: int,
    ):
        tileable = context.tileable
        chunk_input = tileable.cix[chunk_index]

        if index_info.input_axis == 0:
            dtype = chunk_input.index_value.to_pandas().dtype
            index_value = parse_index(
                pd.Index([], dtype=dtype), chunk_input, index, store_data=False
            )
            dtypes = None
        else:
            pd_index = chunk_input.columns_value.to_pandas()
            filtered_index = pd_index[index]
            index_value = parse_index(filtered_index, store_data=True)
            dtypes = chunk_input.dtypes[index]

        info = ChunkIndexAxisInfo(
            output_axis_index=output_axis_index,
            processed_index=index,
            output_shape=output_shape,
            index_value=index_value,
            dtypes=dtypes,
        )
        chunk_index_info.set(info)


class TensorBoolIndexHandler(TensorBoolIndexHandlerBase):
    @classmethod
    def set_chunk_index_info(
        cls,
        context: IndexHandlerContext,
        index_info: IndexInfo,
        chunk_index: Tuple[int],
        chunk_index_info: ChunkIndexInfo,
        output_axis_index: int,
        index,
        output_shape: int,
    ):
        tileable = context.tileable
        chunk_input = tileable.cix[chunk_index]

        assert (
            index_info.input_axis == 0
        ), "bool indexing on axis columns cannot be tensor"

        index_value = parse_index(
            pd.Index([], chunk_input.index_value.to_pandas().dtype),
            chunk_input,
            index,
            store_data=False,
        )

        info = ChunkIndexAxisInfo(
            output_axis_index=output_axis_index,
            processed_index=index,
            output_shape=output_shape,
            index_value=index_value,
            dtypes=None,
        )
        chunk_index_info.set(info)


class _FancyIndexHandler(DataFrameIndexHandler, IndexHandler):
    @classproperty
    def kind(self):  # pylint: disable=no-self-use
        return "iloc"

    def parse(self, raw_index, context: IndexHandlerContext) -> IndexInfo:
        info = FancyIndexInfo(
            IndexType.fancy_index,
            context.input_axis,
            context.output_axis,
            raw_index,
            self,
        )
        context.input_axis += 1
        context.output_axis += 1
        context.append(info)
        return info


class NDArrayFancyIndexHandler(_FancyIndexHandler):
    def accept(cls, raw_index):
        # raw index like list, and pd.Series
        # would have been converted to ndarray or tensor already
        return isinstance(raw_index, np.ndarray) and raw_index.dtype != np.bool_

    def preprocess(self, index_info: IndexInfo, context: IndexHandlerContext) -> None:
        tileable = context.tileable
        if has_unknown_shape(tileable):  # pragma: no cover
            yield

        # split raw index into chunks on the given axis
        split_info = split_indexes_into_chunks(
            [tileable.nsplits[index_info.input_axis]], [index_info.raw_index]
        )
        index_info.split_info = split_info

    def process(self, index_info: IndexInfo, context: IndexHandlerContext) -> None:
        chunk_index_to_fancy_index_arrays = index_info.split_info[0]

        other_index_to_iter = dict()
        chunk_index_to_info = context.chunk_index_to_info.copy()
        for chunk_index, chunk_index_info in chunk_index_to_info.items():
            i = chunk_index[index_info.input_axis]
            fancy_index_array = chunk_index_to_fancy_index_arrays[(i,)][0]

            if fancy_index_array.size == 0:
                # not effected
                del context.chunk_index_to_info[chunk_index]
                continue

            other_index = (
                chunk_index[:1] if index_info.input_axis == 1 else chunk_index[1:]
            )
            if other_index not in other_index_to_iter:
                other_index_to_iter[other_index] = itertools.count()
            output_axis_index = next(other_index_to_iter[other_index])
            output_axis_shape = fancy_index_array.shape[0]
            self.set_chunk_index_info(
                context,
                index_info,
                chunk_index,
                chunk_index_info,
                output_axis_index,
                fancy_index_array,
                output_axis_shape,
            )

    @classmethod
    def need_postprocess(cls, index_info: IndexInfo, context: IndexHandlerContext):
        tileable = context.tileable

        if tileable.chunk_shape[index_info.input_axis] == 1:
            # if tileable only has 1 chunk on this axis
            # do not need postprocess
            return False
        # if ascending sorted, no need to postprocess
        return not index_info.split_info[2]

    def postprocess(self, index_info: IndexInfo, context: IndexHandlerContext) -> None:
        # could be 2 fancy indexes at most
        fancy_indexes = context.get_indexes(index_info.index_type)
        i_fancy_index = fancy_indexes.index(index_info)
        need_postprocesses = [
            fancy_index.handler.need_postprocess(fancy_index, context)
            for fancy_index in fancy_indexes
        ]

        if not need_postprocesses[i_fancy_index]:
            # do not need postprocess
            return

        if (
            i_fancy_index == 0
            and len(fancy_indexes) == 2
            and need_postprocesses[1]
            and isinstance(fancy_indexes[1].raw_index, np.ndarray)
        ):
            # check if need postprocess if 2 fancy indexes and now it's the first,
            # if so, skip postprocess for this one,
            # and do MapReduce just once for the second postprocess
            return

        chunks, nsplits = context.out_chunks, context.out_nsplits
        index_to_chunks = {c.index: c for c in chunks}

        to_concat_axes = tuple(
            fancy_index.output_axis
            for i, fancy_index in enumerate(fancy_indexes)
            if need_postprocesses[i]
        )
        reorder_indexes = [
            calc_pos(fancy_index.raw_index.shape, fancy_index.split_info[1])
            for i, fancy_index in enumerate(fancy_indexes)
            if need_postprocesses[i]
        ]
        new_out_chunks = []
        for chunk_index in itertools.product(
            *(
                range(len(ns))
                for ax, ns in enumerate(nsplits)
                if ax not in to_concat_axes
            )
        ):
            if len(to_concat_axes) == 2:
                to_concat_chunks = chunks
            else:
                to_concat_chunks = []
                for i in range(len(nsplits[to_concat_axes[0]])):
                    to_concat_index = list(chunk_index)
                    to_concat_index.insert(to_concat_axes[0], i)
                    to_concat_chunks.append(index_to_chunks[tuple(to_concat_index)])
            concat_chunk = context.concat_chunks(to_concat_chunks, to_concat_axes)
            reorder_chunk = self._create_reorder_chunk(
                concat_chunk, to_concat_axes, reorder_indexes, context
            )
            new_out_chunks.append(reorder_chunk)

        new_nsplits = list(nsplits)
        for fancy_index in fancy_indexes:
            new_nsplits[fancy_index.output_axis] = (fancy_index.raw_index.shape[0],)
        context.out_chunks = new_out_chunks
        context.out_nsplits = new_nsplits

    @classmethod
    def _create_reorder_chunk(
        cls,
        concat_chunk: Chunk,
        to_concat_axes: Tuple,
        reorder_indexes: List,
        context: IndexHandlerContext,
    ):
        reorder_chunk_op = context.op.copy().reset_key()
        indexes = [slice(None)] * concat_chunk.ndim
        for ax, reorder_index in zip(to_concat_axes, reorder_indexes):
            indexes[ax] = reorder_index
        reorder_chunk_op._indexes = indexes

        params = concat_chunk.params
        if isinstance(concat_chunk, SERIES_CHUNK_TYPE):
            if concat_chunk.index_value.has_value():
                # if concat chunk's index has value, we could calculate the new index
                reorder_index = concat_chunk.index_value.to_pandas()[reorder_indexes[0]]
                params["index_value"] = parse_index(reorder_index, store_data=True)
            else:
                params["index_value"] = parse_index(
                    concat_chunk.index_value.to_pandas(), indexes
                )
            return reorder_chunk_op.new_chunk([concat_chunk], kws=[params])
        else:
            if 0 in to_concat_axes:
                if concat_chunk.index_value.has_value():
                    # if concat chunk's index has value, and index on axis 0,
                    # we could calculate the new index
                    reorder_index = concat_chunk.index_value.to_pandas()[
                        reorder_indexes[0]
                    ]
                    params["index_value"] = parse_index(reorder_index, store_data=True)
                else:
                    params["index_value"] = parse_index(
                        concat_chunk.index_value.to_pandas(), indexes[0]
                    )
            if 1 in to_concat_axes:
                reorder_columns = concat_chunk.columns_value.to_pandas()[
                    reorder_indexes[-1]
                ]
                params["columns_value"] = parse_index(reorder_columns, store_data=True)
                params["dtypes"] = concat_chunk.dtypes[reorder_indexes[-1]]

        return reorder_chunk_op.new_chunk([concat_chunk], kws=[params])


class _LabelFancyIndexHandler(DataFrameIndexHandler, IndexHandler):
    @classproperty
    def kind(self):  # pylint: disable=no-self-use
        return "loc"


class LabelNDArrayFancyIndexHandler(_LabelFancyIndexHandler):
    def accept(cls, raw_index):
        return isinstance(raw_index, np.ndarray) and raw_index.dtype != np.bool_

    def parse(self, raw_index, context: IndexHandlerContext) -> IndexInfo:
        info = LabelFancyIndexInfo(
            IndexType.label_fancy_index,
            context.input_axis,
            context.output_axis,
            raw_index,
            self,
        )
        context.input_axis += 1
        if not np.isscalar(raw_index):
            context.output_axis += 1
        context.append(info)
        return info

    def preprocess(self, index_info: IndexInfo, context: IndexHandlerContext) -> None:
        tileable = context.tileable
        op = context.op

        input_axis = index_info.input_axis

        # check unknown shape
        if any(np.isnan(s) for s in tileable.nsplits[input_axis]):
            yield

        if tileable.ndim == 2:
            index_value = [tileable.index_value, tileable.columns_value][input_axis]
        else:
            index_value = tileable.index_value
        cum_nsplit = [0] + np.cumsum(tileable.nsplits[input_axis]).tolist()
        if not op.can_index_miss and index_value.has_value():
            # df.loc cannot have missed index, reindex can have
            # thus for reindex, do not try to resolve by converting to positions
            # turn label-based fancy index into position-based
            pd_index = index_value.to_pandas()
            positions = convert_labels_into_positions(pd_index, index_info.raw_index)
            split_info = split_indexes_into_chunks(
                [tileable.nsplits[input_axis]], [positions]
            )
            chunk_index_to_pos = split_info[0]
            is_asc_sorted = split_info[-1]

            # convert back to labels for chunk_index
            chunk_index_to_labels = dict()
            for chunk_index, pos in chunk_index_to_pos.items():
                # chunk_index and pos are all list with 1 element
                abs_pos = pos[0] + cum_nsplit[chunk_index[0]]
                if isinstance(pd_index, pd.RangeIndex) and len(abs_pos) == 0:
                    chunk_labels = np.array([], dtype=pd_index.dtype)
                else:
                    chunk_labels = to_numpy(pd_index[abs_pos])
                chunk_index_to_labels[chunk_index[0]] = chunk_labels

            index_info.is_label_asc_sorted = is_asc_sorted
            index_info.chunk_index_to_labels = chunk_index_to_labels
        else:
            index = index_info.raw_index
            if np.isscalar(index):
                # delegation from label index handler
                index = np.atleast_1d(index)
            # does not know the right positions, need postprocess always
            index_info.is_label_asc_sorted = False
            # do df.loc on each chunk
            index_info.chunk_index_to_labels = {
                i: index for i in range(tileable.chunk_shape[input_axis])
            }

    def process(self, index_info: IndexInfo, context: IndexHandlerContext) -> None:
        tileable = context.tileable
        input_axis = index_info.input_axis
        chunk_index_to_labels = index_info.chunk_index_to_labels
        full_label_size = sum(labels.size for labels in chunk_index_to_labels.values())

        other_index_to_iter = dict()
        chunk_index_to_info = context.chunk_index_to_info.copy()
        for chunk_index, chunk_index_info in chunk_index_to_info.items():
            i = chunk_index[input_axis]
            chunk_labels = chunk_index_to_labels[i]
            size = chunk_labels.size

            if size == 0 and full_label_size > 0 and tileable.shape[0] > 0:
                # not effected when
                # 1) tileable not empty
                # 2) full index not empty
                # 3) no index chosen for this chunk
                del context.chunk_index_to_info[chunk_index]
                continue

            if (
                np.isscalar(index_info.raw_index)
                and isinstance(tileable.index_value.value, IndexValue.DatetimeIndex)
                and isinstance(chunk_labels[0], str)
            ):
                # special case when index is DatetimeIndex and loc by string
                # convert back list to scalar because if keep list,
                # KeyError will always happen
                chunk_labels = chunk_labels[0].item()

            other_index = chunk_index[:1] if input_axis == 1 else chunk_index[1:]
            if other_index not in other_index_to_iter:
                other_index_to_iter[other_index] = itertools.count()
            output_axis_index = next(other_index_to_iter[other_index])
            output_axis_shape = size
            self.set_chunk_index_info(
                context,
                index_info,
                chunk_index,
                chunk_index_info,
                output_axis_index,
                chunk_labels,
                output_axis_shape,
            )

    @classmethod
    def need_postprocess(cls, index_info: IndexInfo, context: IndexHandlerContext):
        # if ascending sorted, no need to postprocess
        return not index_info.is_label_asc_sorted

    def postprocess(self, index_info: IndexInfo, context: IndexHandlerContext) -> None:
        if not self.need_postprocess(index_info, context):
            # do not need postprocess
            return

        chunks, nsplits = context.out_chunks, context.out_nsplits
        index_to_chunks = {c.index: c for c in chunks}

        axis = index_info.output_axis
        new_out_chunks = []
        chunk_axis_shapes = dict()
        for chunk_index in itertools.product(
            *(range(len(ns)) for ax, ns in enumerate(nsplits) if ax != axis)
        ):
            to_concat_chunks = []
            for i in range(len(nsplits[axis])):
                if axis == 0:
                    to_concat_index = (i,) + chunk_index
                else:
                    to_concat_index = chunk_index + (i,)
                to_concat_chunks.append(index_to_chunks[to_concat_index])
            concat_chunk = context.concat_chunks(to_concat_chunks, axis)
            chunk_op = context.op.copy().reset_key()
            indexes = [slice(None)] * len(nsplits)
            indexes[axis] = index_info.raw_index
            params = concat_chunk.params
            if np.isscalar(index_info.raw_index):
                assert axis == 0
                if "columns_value" in params:
                    params["index_value"] = params.pop("columns_value")
                    params["dtype"] = find_common_type(params["dtypes"].tolist())
                    del params["dtypes"]
                    if getattr(context.op.outputs[0], "name", None) is not None:
                        params["name"] = context.op.outputs[0].name
                if len(params["index"]) == chunks[0].ndim:
                    index = list(params["index"])
                    index.pop(index_info.output_axis)
                    params["index"] = tuple(index)
                    shape = list(params["shape"])
                    shape.pop(index_info.output_axis)
                    params["shape"] = tuple(shape)
                if context.op.outputs[0].ndim == 0:
                    del params["index_value"]
            elif axis == 0:
                pd_index = pd.Index(index_info.raw_index)
                params["index_value"] = parse_index(pd_index, store_data=False)
                shape = list(params["shape"])
                shape[0] = len(pd_index)
                params["shape"] = shape
            else:
                if context.op.can_index_miss:
                    # reindex
                    params["dtypes"] = dtypes = to_concat_chunks[0].dtypes
                else:
                    params["dtypes"] = dtypes = concat_chunk.dtypes.loc[
                        index_info.raw_index
                    ]
                params["columns_value"] = parse_index(dtypes.index, store_data=True)
                shape = list(params["shape"])
                shape[1] = len(dtypes)
                params["shape"] = tuple(shape)
            chunk_op._indexes = indexes
            chunk_op.stage = OperandStage.agg
            out_chunk = chunk_op.new_chunk([concat_chunk], kws=[params])
            if len(out_chunk.shape) != 0:
                chunk_axis_shapes[out_chunk.index[axis]] = out_chunk.shape[axis]
            new_out_chunks.append(out_chunk)

        new_nsplits = list(nsplits)
        if np.isscalar(index_info.raw_index):
            new_nsplits = new_nsplits[:axis] + new_nsplits[axis + 1 :]
        else:
            new_nsplits[axis] = (sum(chunk_axis_shapes.values()),)
        context.out_chunks = new_out_chunks
        context.out_nsplits = new_nsplits


class LabelTensorFancyIndexHandler(_LabelFancyIndexHandler):
    def accept(cls, raw_index):
        return isinstance(raw_index, TENSOR_TYPE) and raw_index.dtype != np.bool_

    def parse(self, raw_index, context: IndexHandlerContext) -> IndexInfo:
        if context.input_axis == 1:  # pragma: no cover
            raise NotImplementedError(
                "do not support tensor-based index on columns axis"
            )
        info = LabelFancyIndexInfo(
            IndexType.label_fancy_index,
            context.input_axis,
            context.output_axis,
            raw_index,
            self,
        )
        context.input_axis += 1
        context.output_axis += 1
        context.append(info)
        return info

    def preprocess(self, index_info: IndexInfo, context: IndexHandlerContext) -> None:
        if has_unknown_shape(index_info.raw_index):
            yield
        # rechunk index into one
        index_info.unprocessed_raw_index = index_info.raw_index
        index_info.raw_index = yield from recursive_tile(
            index_info.raw_index.rechunk(index_info.raw_index.shape)
        )

    def process(self, index_info: IndexInfo, context: IndexHandlerContext) -> None:
        tileable = context.tileable
        input_axis = index_info.input_axis

        assert len(index_info.raw_index.chunks) == 1
        chunk_labels = index_info.raw_index.chunks[0]

        other_index_to_iter = dict()
        for chunk in tileable.chunks:
            chunk_index = chunk.index
            other_index = chunk_index[:1] if input_axis == 1 else chunk_index[1:]
            if other_index not in other_index_to_iter:
                other_index_to_iter[other_index] = itertools.count()
            output_axis_index = next(other_index_to_iter[other_index])
            self.set_chunk_index_info(
                context,
                index_info,
                chunk_index,
                context.chunk_index_to_info[chunk.index],
                output_axis_index,
                chunk_labels,
                len(chunk_labels),
            )

    def postprocess(self, index_info: IndexInfo, context: IndexHandlerContext) -> None:
        from .iloc import DataFrameIlocGetItem

        tileable = context.tileable
        out = context.op.outputs[0]
        max_chunk_size = max(tileable.nsplits[index_info.input_axis])
        max_chunk_size = 0 if np.isnan(max_chunk_size) else max_chunk_size
        max_chunk_size = max(
            max_chunk_size, max(index_info.unprocessed_raw_index.nsplits[0])
        )
        new_chunk_sizes = normalize_chunk_sizes(
            index_info.raw_index.shape[0], max_chunk_size
        )[0]
        cum_new_chunk_sizes = [0] + np.cumsum(new_chunk_sizes).tolist()

        chunks, nsplits = context.out_chunks, context.out_nsplits
        index_to_chunks = {c.index: c for c in chunks}

        axis = index_info.output_axis
        assert axis == 0
        new_out_chunks = []
        for chunk_index in itertools.product(
            *(range(len(ns)) for ax, ns in enumerate(nsplits) if ax != axis)
        ):
            select_chunks = []
            for i in range(len(nsplits[axis])):
                select_index = (i,) + chunk_index
                select_chunks.append(index_to_chunks[select_index])

            for j in range(len(new_chunk_sizes)):
                slc = slice(cum_new_chunk_sizes[j], cum_new_chunk_sizes[j + 1])
                indexes = [slice(None)] * len(nsplits)
                indexes[axis] = slc

                slice_chunks = []
                for select_chunk in select_chunks:
                    output_types = (
                        [OutputType.series]
                        if len(nsplits) == 1
                        else [OutputType.dataframe]
                    )
                    slc_op = DataFrameIlocGetItem(
                        indexes=indexes, output_types=output_types
                    )
                    slice_chunk_shape = list(select_chunk.shape)
                    slice_chunk_shape[axis] = new_chunk_sizes[j]
                    slice_chunk = slc_op.new_chunk(
                        [select_chunk], shape=tuple(slice_chunk_shape)
                    )
                    slice_chunks.append(slice_chunk)

                chunk_op = context.op.copy().reset_key()
                chunk_op.stage = OperandStage.agg
                chunk_op._indexes = (None,) * len(nsplits)
                chunk_op._fill_value = None
                assert axis == 0
                params = dict()
                params["index"] = (
                    (j,) + chunk_index if axis == 0 else chunk_index + (j,)
                )
                params["index_value"] = parse_index(
                    out.index_value.to_pandas()[slc], slice_chunks, store_data=False
                )
                if select_chunks[0].ndim == 2:
                    params["columns_value"] = select_chunks[0].columns_value
                    params["dtypes"] = select_chunks[0].dtypes
                else:
                    params["dtype"] = select_chunks[0].dtype
                    params["name"] = select_chunks[0].name
                params["shape"] = slice_chunks[0].shape
                out_chunk = chunk_op.new_chunk(slice_chunks, kws=[params])
                new_out_chunks.append(out_chunk)

        new_nsplits = list(nsplits)
        new_nsplits[axis] = tuple(new_chunk_sizes)
        context.out_chunks = new_out_chunks
        context.out_nsplits = new_nsplits


class DataFrameIlocIndexesHandler(IndexesHandler):
    def __init__(self):
        super().__init__()
        self.register(
            IntegralIndexHandler,
            SliceIndexHandler,
            NDArrayBoolIndexHandler,
            TensorBoolIndexHandler,
            NDArrayFancyIndexHandler,
        )

    def create_context(self, op):
        return DataFrameIndexHandlerContext(op)


class DataFrameLocIndexesHandler(IndexesHandler):
    def __init__(self):
        super().__init__()
        self.register(
            LabelIndexHandler,
            LabelSliceIndexHandler,
            NDArrayBoolIndexHandler,
            TensorBoolIndexHandler,
            LabelNDArrayFancyIndexHandler,
        )

    def create_context(self, op):
        return DataFrameIndexHandlerContext(op)


class DataFrameReindexHandler(IndexesHandler):
    def __init__(self):
        super().__init__()
        self.register(
            LabelTensorFancyIndexHandler,
            LabelNDArrayFancyIndexHandler,
            LabelSliceIndexHandler,
        )

    def create_context(self, op):
        return DataFrameIndexHandlerContext(op)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect
import logging
from dataclasses import asdict
from typing import Dict, List, Callable

from .....core.context import Context
from .....storage.base import StorageLevel
from .....utils import implements, lazy_import
from ....context import ThreadedServiceContext
from .config import RayExecutionConfig

ray = lazy_import("ray")
logger = logging.getLogger(__name__)


class RayRemoteObjectManager:
    """The remote object manager in task state actor."""

    def __init__(self):
        self._named_remote_objects = {}

    def create_remote_object(self, name: str, object_cls, *args, **kwargs):
        remote_object = object_cls(*args, **kwargs)
        self._named_remote_objects[name] = remote_object

    def destroy_remote_object(self, name: str):
        self._named_remote_objects.pop(name, None)

    async def call_remote_object(self, name: str, attr: str, *args, **kwargs):
        remote_object = self._named_remote_objects[name]
        meth = getattr(remote_object, attr)
        async_meth = self._sync_to_async(meth)
        return await async_meth(*args, **kwargs)

    @staticmethod
    @functools.lru_cache(100)
    def _sync_to_async(func):
        if inspect.iscoroutinefunction(func):
            return func
        else:

            async def async_wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return async_wrapper


class _RayRemoteObjectWrapper:
    def __init__(self, task_state_actor: "ray.actor.ActorHandle", name: str):
        self._task_state_actor = task_state_actor
        self._name = name

    def __getattr__(self, attr):
        def wrap(*args, **kwargs):
            r = self._task_state_actor.call_remote_object.remote(
                self._name, attr, *args, **kwargs
            )
            return ray.get(r)

        return wrap


class _RayRemoteObjectContext:
    def __init__(
        self,
        get_or_create_actor: Callable[[], "ray.actor.ActorHandle"],
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._get_or_create_actor = get_or_create_actor
        self._task_state_actor = None

    def _get_task_state_actor(self) -> "ray.actor.ActorHandle":
        # Get the RayTaskState actor, this is more clear and faster than wraps
        # the `get_or_create_actor` by lru_cache in __init__ because this method
        # is called as needed.
        if self._task_state_actor is None:
            self._task_state_actor = self._get_or_create_actor()
        return self._task_state_actor

    @implements(Context.create_remote_object)
    def create_remote_object(self, name: str, object_cls, *args, **kwargs):
        task_state_actor = self._get_task_state_actor()
        r = task_state_actor.create_remote_object.remote(
            name, object_cls, *args, **kwargs
        )
        # Make sure the actor is created. The remote object may not be created
        # when get_remote_object from worker because the callers of
        # create_remote_object and get_remote_object are not in the same worker.
        # Use sync Ray actor requires this `ray.get`, too.
        ray.get(r)
        return _RayRemoteObjectWrapper(task_state_actor, name)

    @implements(Context.get_remote_object)
    def get_remote_object(self, name: str):
        task_state_actor = self._get_task_state_actor()
        return _RayRemoteObjectWrapper(task_state_actor, name)

    @implements(Context.destroy_remote_object)
    def destroy_remote_object(self, name: str):
        task_state_actor = self._get_task_state_actor()
        task_state_actor.destroy_remote_object.remote(name)


# TODO(fyrestone): Implement more APIs for Ray.
class RayExecutionContext(_RayRemoteObjectContext, ThreadedServiceContext):
    """The context for tiling."""

    def __init__(
        self,
        config: RayExecutionConfig,
        task_context: Dict,
        task_chunks_meta: Dict,
        worker_addresses: List[str],
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._config = config
        self._task_context = task_context
        self._task_chunks_meta = task_chunks_meta
        self._worker_addresses = worker_addresses

    @implements(Context.get_chunks_result)
    def get_chunks_result(self, data_keys: List[str], fetch_only: bool = False) -> List:
        logger.info("Getting %s chunks result.", len(data_keys))
        object_refs = [self._task_context[key] for key in data_keys]
        result = ray.get(object_refs)
        logger.info("Got %s chunks result.", len(result))
        return result if not fetch_only else None

    @implements(Context.get_chunks_meta)
    def get_chunks_meta(
        self, data_keys: List[str], fields: List[str] = None, error="raise"
    ) -> List[Dict]:
        result = []
        # TODO(fyrestone): Support get_chunks_meta from meta service if needed.
        for key in data_keys:
            chunk_meta = self._task_chunks_meta[key]
            meta = asdict(chunk_meta)
            meta = {f: meta.get(f) for f in fields}
            result.append(meta)
        return result

    @implements(Context.get_total_n_cpu)
    def get_total_n_cpu(self) -> int:
        # TODO(fyrestone): Support auto scaling.
        return self._config.get_n_cpu() * self._config.get_n_worker()

    @implements(Context.get_worker_addresses)
    def get_worker_addresses(self) -> List[str]:
        # Returns virtual worker addresses.
        return self._worker_addresses


# TODO(fyrestone): Implement more APIs for Ray.
class RayExecutionWorkerContext(_RayRemoteObjectContext, dict):
    """The context for executing operands."""

    @classmethod
    @implements(Context.new_custom_log_dir)
    def new_custom_log_dir(cls):
        logger.info(
            "%s does not support register_custom_log_path / new_custom_log_dir",
            cls.__name__,
        )
        return None

    @staticmethod
    @implements(Context.register_custom_log_path)
    def register_custom_log_path(
        session_id: str,
        tileable_op_key: str,
        chunk_op_key: str,
        worker_address: str,
        log_path: str,
    ):
        raise NotImplementedError

    @classmethod
    @implements(Context.set_progress)
    def set_progress(cls, progress: float):
        logger.info(
            "%s does not support set_running_operand_key / set_progress", cls.__name__
        )

    @staticmethod
    @implements(Context.set_running_operand_key)
    def set_running_operand_key(session_id: str, op_key: str):
        raise NotImplementedError

    @classmethod
    @implements(Context.get_storage_info)
    def get_storage_info(
        cls, address: str = None, level: StorageLevel = StorageLevel.MEMORY
    ):
        logger.info("%s does not support get_storage_info", cls.__name__)
        return {}
